#include "TComDict.h"
#include "TComPicYuv.h"
#include "TComRom.h"
#include "TComDataCU.h"
#include "TComPic.h"
#include "TComTrQuant.h"
#include <cassert>

#if CE10_INTRA_STRING_COPY
Bool operator == (const TCombinePixel& _left, const TCombinePixel& _right)
{
	return (_left.Y == _right.Y) && (_left.U == _right.U) && (_left.V == _right.V);
}

Bool operator != (const TCombinePixel& _left, const TCombinePixel& _right)
{
	return !(_left == _right);
}

TComPseudoDict::TComPseudoDict( Int picWidth, Int picHeight,  UInt uiMaxCUWidth, UInt uiMaxCUHeight,  Bool isEncoding )
	//: m_crcCalculator( 16, 0x8005 )
	//: m_crcCalculator( 8, 0x31 )
	: m_crcCalculator( 12, 0x80B )
{
	m_uiMaxCuWidth = uiMaxCUWidth;
	m_uiMaxCuHeight= uiMaxCUHeight;

#if ISC_WPP_STEP
	m_uiCuWidthBits  = (UInt)(log10((Double)uiMaxCUWidth)/log10(2.0));
	m_uiCuHeightBits = (UInt)(log10((Double)uiMaxCUHeight)/log10(2.0));
#endif
	m_picWidth = picWidth;
	m_picHeight = picHeight;
	m_isEncoding = isEncoding;
	if ( m_isEncoding )
	{
		m_dict = new TDictPreviousMatchPosition*[m_picHeight];
		for ( Int i=0; i<m_picHeight; i++ )
		{
			m_dict[i] = new TDictPreviousMatchPosition[m_picWidth];
		}
	}
	else
	{
		m_dict = NULL;
	}
	for ( Int i=0; i<3; i++ )
	{
		m_pPelOrg[i] = NULL;
		m_pPelRec[i] = NULL;
	}

	m_pCRCCalculationBuffer = new UChar[3];
	reset( NULL, NULL );
}

TComPseudoDict::~TComPseudoDict()
{
	if ( m_isEncoding )
	{
		if ( m_dict != NULL )
		{
			for ( Int i=0; i<m_picHeight; i++ )
			{
				if ( m_dict[i] != NULL )
				{
					delete[] m_dict[i];
					m_dict[i] = NULL;
				}
			}
			delete[] m_dict;
			m_dict = NULL;
		}
	}

	if ( m_pCRCCalculationBuffer != NULL )
	{
		delete[] m_pCRCCalculationBuffer;
		m_pCRCCalculationBuffer = NULL;
	}
}

Void TComPseudoDict::reset( TComPicYuv* pOrg, TComPicYuv* pRec )
{
	for ( Int i=0; i<3; i++ )
	{
		ComponentID compID = ComponentID( i );
		if ( pOrg != NULL )
		{
			m_pPelOrg[i] = pOrg->getAddr( compID );
			m_strideOrg[i] = pOrg->getStride( compID );
		}
		if ( pRec != NULL )
		{
			m_pPelRec[i] = pRec->getAddr( compID );
			m_strideRec[i] = pRec->getStride( compID );
		}
	}
	if ( m_isEncoding )
	{
		for ( Int i=0; i<DICT_PSEUDO_2D_HASH_SIZE; i++ )
		{
			m_hashTable[i].x = -1;
			m_hashTable[i].y = -1;
		}
	}
	if (g_isLossless)
	{
		getCRCValue = &TComPseudoDict::getCRCValueLossless;
	} 
	else
	{
		getCRCValue = &TComPseudoDict::getCRCValueLossy;
	}
}

Void TComPseudoDict::addToDict( Int xStartInPic, Int yStartInPic, Int width, Int height, UInt uiScaleX, UInt uiScaleY )
{
	TCombinePixel CurPixel;
	for ( Int i=0; i<height; i++ )
	{
		for ( Int j=0; j<width; j++ )
		{
			Int currentX = xStartInPic+j;
			Int currentY = yStartInPic+i;
			//Int length = 1;
			Int lengthMinus1 = 0;
			Int startPositionX = currentX - lengthMinus1;
			Int startPositionY = currentY;
			if ( startPositionX < 0 || startPositionY < 0 )
			{
				continue;
			}
      Int curPos = currentY * m_strideOrg[0] + currentX;
			Int curPosC= (currentY>>uiScaleY) * m_strideOrg[1] + (currentX>>uiScaleX);
			CurPixel.Y = m_pPelOrg[0][curPos ];
			CurPixel.U = m_pPelOrg[1][curPosC];
			CurPixel.V = m_pPelOrg[2][curPosC];
			UShort hashValue = (this->*getCRCValue)(CurPixel);

			m_dict[currentY][currentX].x = m_hashTable[hashValue].x;
			m_dict[currentY][currentX].y = m_hashTable[hashValue].y;
			m_hashTable[hashValue].x = startPositionX;
			m_hashTable[hashValue].y = startPositionY;
		}
	}
}

UShort TComPseudoDict::xGetCRCValue( UChar* p, Int length )
{
	m_crcCalculator.reset();
	m_crcCalculator.processData( p, sizeof(UChar)* 3 * length );
	return m_crcCalculator.getCRC();
}

UShort TComPseudoDict::getCRCValueLossy(TCombinePixel pixelCur) 
{
	UChar* p = m_pCRCCalculationBuffer;
	p[0] = static_cast<UChar>(pixelCur.Y);//static_cast<UChar>(m_pPelOrg[0][yEnd*m_strideOrg[0] + xEnd ]);
	p[1] = static_cast<UChar>(pixelCur.U);//static_cast<UChar>(m_pPelOrg[1][yEnd*m_strideOrg[1] + xEnd ]);
	p[2] = static_cast<UChar>(pixelCur.V);//static_cast<UChar>(m_pPelOrg[2][yEnd*m_strideOrg[2] + xEnd ]);
	
	UShort crc = ((p[0] & 0xf0)<<4) | ((p[1] & 0xf0)) | ((p[2] & 0xf0)>>4);

	return crc;
}
UShort TComPseudoDict::getCRCValueLossless(TCombinePixel pixelCur) 
{
	UChar* p = m_pCRCCalculationBuffer;
	p[0] = static_cast<UChar>(pixelCur.Y);//static_cast<UChar>(m_pPelOrg[0][yEnd*m_strideOrg[0] + xEnd ]);
	p[1] = static_cast<UChar>(pixelCur.U);//static_cast<UChar>(m_pPelOrg[1][yEnd*m_strideOrg[1] + xEnd ]);
	p[2] = static_cast<UChar>(pixelCur.V);//static_cast<UChar>(m_pPelOrg[2][yEnd*m_strideOrg[2] + xEnd ]);
  //
	UShort crc = (xGetCRCValue( p, 1 ));
	//crc &= (DICT_PSEUDO_2D_HASH_SIZE-1);

	return crc;
}

Void TComPseudoDict::save(TDictStatus* pDictStatus, Int xLCUStart, Int yLCUStart, Int width, Int height)
{
	memcpy(pDictStatus->m_hashTable, m_hashTable, sizeof(TDictPreviousMatchPosition)*DICT_PSEUDO_2D_HASH_SIZE);
	Int stride = (xLCUStart + MAX_CU_SIZE) > m_picWidth ? (m_picWidth - xLCUStart) : MAX_CU_SIZE;
	for (Int j = 0; j < MAX_CU_SIZE; j++)
	{
		if ( (yLCUStart + j) < m_picHeight )
		{
			memcpy( pDictStatus->m_dict[j], &m_dict[yLCUStart + j][xLCUStart], sizeof( TDictPreviousMatchPosition )*stride );
		}
	}
}

Void TComPseudoDict::restore(TDictStatus* pDictStatus, Int xLCUStart, Int yLCUStart, Int width, Int height)
{
	memcpy(m_hashTable, pDictStatus->m_hashTable, sizeof(TDictPreviousMatchPosition)*DICT_PSEUDO_2D_HASH_SIZE);
	Int stride = (xLCUStart + MAX_CU_SIZE) > m_picWidth ? (m_picWidth - xLCUStart) : MAX_CU_SIZE;
	for (Int j = 0; j < MAX_CU_SIZE; j++)
	{
		if ( (yLCUStart + j) < m_picHeight )
		{
			memcpy( &m_dict[yLCUStart + j][xLCUStart], pDictStatus->m_dict[j], sizeof( TDictPreviousMatchPosition )* stride );
		}
	}
}

Void TDictStatus::save(TDictStatus* p)
{
	memcpy(p->m_hashTable, m_hashTable, sizeof(TDictPreviousMatchPosition)*DICT_PSEUDO_2D_HASH_SIZE);
	for (Int j = 0; j < MAX_CU_SIZE; j++)
	{
		memcpy(p->m_dict[j], m_dict[j], sizeof(TDictPreviousMatchPosition)*MAX_CU_SIZE);//ZF
	}
	p->m_currDictSize = m_currDictSize;
}

Void TComPseudoDict::SRBPredition(TComDataCU* pcCU, TCombinePixel *pHiRefUmP, UInt uiHiRefUmPSize, UChar* pReusedPrev)
{
	Pel *pPred[3]   = { pcCU->getLastPLTInLcuFinal(0), pcCU->getLastPLTInLcuFinal(1), pcCU->getLastPLTInLcuFinal(2) };
	UInt uiPLTSizePrev = pcCU->getLastPLTInLcuSizeFinal(0);
	//set SRB Reused Flag
	Pel pSRBTemp[3][MAX_SRB_SIZE];
	UInt uiIdxPrev = 0, uiIdxCurr = 0;
	Bool bReused;
	Bool pPredicted[MAX_SRB_SIZE + 1];    
	memset(pPredicted, 0, sizeof(pPredicted));

	//save cur SRB into temp
	for (UInt i = 0; i < MAX_SRB_SIZE; i++)
	{
		pSRBTemp[0][i] = pHiRefUmP[i].Y;
		pSRBTemp[1][i] = pHiRefUmP[i].U;
		pSRBTemp[2][i] = pHiRefUmP[i].V;
	}
	//
	for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev++)
	{
		bReused = false;
		Int iCounter = 0;
		for (uiIdxCurr = 0; uiIdxCurr < uiHiRefUmPSize; uiIdxCurr++)
		{
			iCounter = 0;

			for (UInt ch = 0; ch < 3; ch++)
			{
				if (pPred[ch][uiIdxPrev] == pSRBTemp[ch][uiIdxCurr])
				{
					iCounter++;
				}
			}
			if (iCounter == 3)
			{
				bReused = true;
				break;
			}
		}
		pReusedPrev[uiIdxPrev] = bReused; 
		pPredicted[uiIdxCurr] = bReused;
	}

	//reorder SRB
	uiIdxCurr = 0;
	for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev++)
	{
		if (pReusedPrev[uiIdxPrev])
		{      
			pHiRefUmP[uiIdxCurr].Y=(UChar) pPred[0][uiIdxPrev];
			pHiRefUmP[uiIdxCurr].U=(UChar) pPred[1][uiIdxPrev];
			pHiRefUmP[uiIdxCurr].V=(UChar) pPred[2][uiIdxPrev];
			uiIdxCurr++;
		}
	}

	for (UInt uiIdx = 0; uiIdx < uiHiRefUmPSize; uiIdx++)
	{
		if (pPredicted[uiIdx] == 0)
		{      
			pHiRefUmP[uiIdxCurr].Y=(UChar) pSRBTemp[0][uiIdx];
			pHiRefUmP[uiIdxCurr].U=(UChar) pSRBTemp[1][uiIdx];
			pHiRefUmP[uiIdxCurr].V=(UChar) pSRBTemp[2][uiIdx];
			uiIdxCurr++;
		}
	}
}

Void TComPseudoDict::saveSRB(TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth)
{
	if(m_bHiRefFlag)
		pcCU->setSRBFlagSubParts(true, uiAbsPartIdx, uiDepth);

	pcCU->setSRBIscFlagSubParts(false, uiAbsPartIdx, uiDepth);
	pcCU->setSRBLoRefFlagSubParts(0, m_bLoRefFlag, uiAbsPartIdx, uiDepth);
	//set SRB
	assert(m_uiHiRefUmPSize<=m_MaxPLTSize);
	for(UInt com=0;com<3;com++)
	{
		pcCU->setSRBSizeSubParts(com,m_uiHiRefUmPSize,0,uiDepth);
	}

	UChar *pReusedPrev=new UChar[MAX_PLT_PRED_SIZE + 1];
	memset(pReusedPrev, false, sizeof(UChar)*(MAX_PLT_PRED_SIZE + 1));
	{
		//SRB prediction
		SRBPredition(pcCU, m_HiRefUmPixel, m_uiHiRefUmPSize, pReusedPrev);
	}

	for (UInt uiIdx = 0; uiIdx < m_PLTMaxPredSize; uiIdx++)
	{
		for (UInt ch = 0; ch < 3; ch++)
		{
			pcCU->setPrevSRBReusedFlagSubParts(ch, pReusedPrev[uiIdx], uiIdx, 0, uiDepth);
		}
	}
	for (UInt uiIdx = 0; uiIdx < m_MaxPLTSize; uiIdx++)
	{
		pcCU->setSRBSubParts(0, m_HiRefUmPixel[uiIdx].Y, uiIdx, 0, uiDepth);
		pcCU->setSRBSubParts(1, m_HiRefUmPixel[uiIdx].U, uiIdx, 0, uiDepth);
		pcCU->setSRBSubParts(2, m_HiRefUmPixel[uiIdx].V, uiIdx, 0, uiDepth);
	}

	if(pReusedPrev)
		delete[] pReusedPrev;
}

Void TComPseudoDict::deriveCluster(TComDataCU* pcCU, UInt uiWidth, UInt uiHeight, Int xStartInPic, Int yStartInPic, Bool bLosslessMode, Double dLambda)
{
	UInt uiClusterSize = m_PredFlag ? g_uiClusterSizePred : g_uiClusterSize;
	Pel *pCluster[3];
	TCombinePixel CurPixel;
  Int  iErrorLimit = bLosslessMode ? 0 : g_uhPLTQuant[Int(pcCU->getQP(0))];  
	m_uiHiRefUmPSize=0;
	for (UInt ch = 0; ch < 3; ch++)
	{
    //get cluster from palette
		pCluster[ch] = m_PredFlag ? g_pClusterPred[ch] : g_pCluster[ch];
	}
	UInt uiPos;
	UInt uiScaleX = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);
	UInt uiScaleY = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
	//gen major flag;  
	UInt  uiTotal = uiWidth * uiHeight;
	Bool* matchFlag =new Bool[uiTotal];  
	UInt CuMatchCnt =0;
	UInt uiBestIdx = 0;
	uiPos=0;
	for (UInt uiY = 0; uiY < uiHeight; uiY++)
	{
		for (UInt uiX = 0; uiX < uiWidth; uiX++)
		{
			Int currentX = xStartInPic + uiX;
			Int currentY = yStartInPic + uiY;
			Int curPos = currentY * m_strideOrg[0] + currentX;
			Int curPosC= (currentY>>uiScaleY) * m_strideOrg[1] + (currentX>>uiScaleX);
			CurPixel.Y = m_pPelOrg[0][curPos ];
			CurPixel.U = m_pPelOrg[1][curPosC];
			CurPixel.V = m_pPelOrg[2][curPosC];
			//save org pixel
			m_pOrgPixelBuffer[uiPos].Y = CurPixel.Y;
			m_pOrgPixelBuffer[uiPos].U = CurPixel.U;
			m_pOrgPixelBuffer[uiPos].V = CurPixel.V;
			//
			Int i = 0;
			UInt uiMinError = MAX_UINT;      
			while (i < uiClusterSize)
			{
				UInt uiAbsError = abs(pCluster[0][i] - CurPixel.Y) + abs(pCluster[1][i] - CurPixel.U) + abs(pCluster[2][i] - CurPixel.V);
				if (uiAbsError < uiMinError)
				{
					uiBestIdx = i;
					uiMinError = uiAbsError;
					if (uiMinError == 0)
					{
						break;
					}
				}
				i++;
			}
			Bool bMismatch = uiMinError > 3 * iErrorLimit;
			matchFlag[uiPos] = bMismatch;
			m_bMajor[uiPos] = true;
			m_cMatchIndex[uiPos] = uiBestIdx;
			if(bMismatch)
			{
				m_bMajor[uiPos]=false;
				m_cMatchIndex[uiPos] = 0xff;
			}
			uiPos++;
		}
	}
	//gen CQ Pixel buffer
  for(UInt i=0;i<uiTotal;i++)
	{
		if(!matchFlag[i])
		{
			m_pCQPixelBuffer[i].Y =  pCluster[0][m_cMatchIndex[i]];
			m_pCQPixelBuffer[i].U =  pCluster[1][m_cMatchIndex[i]];
			m_pCQPixelBuffer[i].V =  pCluster[2][m_cMatchIndex[i]];      
			CuMatchCnt++;
		}
		else
		{			
			m_pCQPixelBuffer[i].Y = m_pOrgPixelBuffer[i].Y;
			m_pCQPixelBuffer[i].U = m_pOrgPixelBuffer[i].U;
			m_pCQPixelBuffer[i].V = m_pOrgPixelBuffer[i].V;			
		}
	}
	//
	m_uiSRBSize = uiClusterSize;

	if(matchFlag)
	{
		delete[] matchFlag;
		matchFlag=NULL;
	}
}

Int TComPseudoDict::getSrbLength(TComDataCU*pcCU, Int width, Int height, Int xCuStartInPic, Int yCuStartInPic, Int offsetInCu, UInt& dist, TCombinePixel& pixelCur)
{
	Int srbLength = 0;
	UInt uiTotal = width * height;
	Bool isHorizontalScanning = pcCU->getDictStatus( 0 ) == DICT_STATUS_REC_HORIZONTAL;
	Int  CbSize    = width;
	Int  log2CbSize=g_aucConvertToBit[width]  + 2;
	TCombinePixel pixelOrgCur;
	TCombinePixel pixelOrg;
	TCombinePixel pixelNxt;
	UInt* puiScanOrder = g_scanOrder[SCAN_UNGROUPED][SCAN_TRAV][g_aucConvertToBit[width] + 2][g_aucConvertToBit[height] + 2];
	UInt uiTraIdx = puiScanOrder[offsetInCu];
	Int  cntNxt = offsetInCu + 1;
	Bool bMajor;
	Int curPosX;
	Int curPosY;
	//Int errorLimit = pcCU->getCUTransquantBypass(0) ? 1 : (1 + 2);

	if(!isHorizontalScanning)
	{
		curPosX = getTravY(uiTraIdx, log2CbSize);//uiTraIdx / height;
		curPosY = getTravX(uiTraIdx, CbSize    );//uiTraIdx % height;
		uiTraIdx= (curPosX) + (curPosY) * width;
	}

	pixelOrgCur= pixelOrg = m_pOrgPixelBuffer[uiTraIdx];
	pixelCur = m_pCQPixelBuffer[uiTraIdx];
	bMajor   = m_bMajor[uiTraIdx];

	dist = CalcDist(pixelOrg, pixelCur);

	if(!bMajor)
		srbLength = -1;
	else
	{
		for (srbLength = 1; cntNxt < uiTotal; srbLength++)
		{
			uiTraIdx = puiScanOrder[cntNxt];
			if(!isHorizontalScanning)
			{
				curPosX = getTravY(uiTraIdx, log2CbSize);//uiTraIdx / height;
				curPosY = getTravX(uiTraIdx, CbSize    );//uiTraIdx % height;
				uiTraIdx= (curPosX) + (curPosY) * width;
			}
			//
			pixelOrg = m_pOrgPixelBuffer[uiTraIdx];
			pixelNxt = m_pCQPixelBuffer[uiTraIdx];
			if(pixelCur == pixelNxt)
			{
				dist += CalcDist(pixelOrg, pixelCur);
				cntNxt++;				
			}
			else
				break;			
		}
	}
	return srbLength;
}

Int TComPseudoDict::getSrbAboveLength(TComDataCU*pcCU, Int width, Int height, Int xCuStartInPic, Int yCuStartInPic, Int offsetInCu, UInt& dist, TCombinePixel& pixelCur, Bool bUseOrg)
{
	Int aboveLength = 0;
	UInt uiTotal = width * height;
	Bool isHorizontalScanning = pcCU->getDictStatus( 0 ) == DICT_STATUS_REC_HORIZONTAL;
	Int  CbSize    = width;
	Int  log2CbSize=g_aucConvertToBit[width]  + 2;
	Int  lossyMatchErr = bUseOrg ? (1 + g_uhPLTQuant[Int(pcCU->getQP(0))]) : 1;
	Int  errorLimit = pcCU->getCUTransquantBypass(0) ? 1 : lossyMatchErr;
	TCombinePixel pixelOrg;
	TCombinePixel pixelRec;
	TCombinePixel pixelAbv;
	UInt uiTraIdx;
	UInt uiTraIdxAbv;
	UInt* puiScanOrder = g_scanOrder[SCAN_UNGROUPED][SCAN_TRAV][g_aucConvertToBit[width] + 2][g_aucConvertToBit[height] + 2];
	UInt uiIdxTrav;
	UInt uiInverseOrder;
	Int  cntNxt = offsetInCu + 1;
	Int  offsetXInCu;
	Int  offsetYInCu;
	Int  curPosX;
	Int  curPosY;
	if( offsetInCu < width )
		return -1;

	TCombinePixel *pRecLineBuf=new TCombinePixel[MAX_CU_SIZE];//line buf in cu;
	UInt uiScaleX = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);
	UInt uiScaleY = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
	uiIdxTrav = puiScanOrder[offsetInCu];

	if(isHorizontalScanning)
	{
		offsetXInCu = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
		offsetYInCu = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;
		curPosX = xCuStartInPic + offsetXInCu;
		curPosY = yCuStartInPic + offsetYInCu;
		//
		uiTraIdx = uiIdxTrav;
		uiTraIdxAbv = uiTraIdx - width;
		uiInverseOrder=offsetYInCu & 0x1;
		//store line buf
		for(Int i=0; i<width; i++)
		{
			if( (i < offsetXInCu && 0 == uiInverseOrder) ||  (i > offsetXInCu && 1 == uiInverseOrder) )
			{
				pRecLineBuf[i].Y=m_pPelRec[0][curPosY*m_strideRec[0] + xCuStartInPic+i];
				pRecLineBuf[i].U=m_pPelRec[1][(curPosY>>uiScaleY)*m_strideRec[1] + ((xCuStartInPic+i)>>uiScaleX)];
				pRecLineBuf[i].V=m_pPelRec[2][(curPosY>>uiScaleY)*m_strideRec[2] + ((xCuStartInPic+i)>>uiScaleX)];
			}
			else
			{
				pRecLineBuf[i].Y=m_pPelRec[0][(curPosY-1)*m_strideRec[0] + xCuStartInPic+i];
				pRecLineBuf[i].U=m_pPelRec[1][((curPosY-1)>>uiScaleY)*m_strideRec[1] + ((xCuStartInPic+i)>>uiScaleX)];
				pRecLineBuf[i].V=m_pPelRec[2][((curPosY-1)>>uiScaleY)*m_strideRec[2] + ((xCuStartInPic+i)>>uiScaleX)];
			}
		}
	}
	else
	{
		offsetXInCu = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
		offsetYInCu = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
		curPosX = xCuStartInPic + offsetXInCu;
		curPosY = yCuStartInPic + offsetYInCu;
		//
		uiTraIdx = offsetXInCu + offsetYInCu * width;
		uiTraIdxAbv = uiTraIdx - 1;
		uiInverseOrder=offsetXInCu & 0x1;
		//store line buf
		for(Int i=0; i<width; i++)
		{
			if( (i < offsetYInCu && 0==uiInverseOrder) || (i > offsetYInCu && 1==uiInverseOrder) )
			{
				pRecLineBuf[i].Y=m_pPelRec[0][(yCuStartInPic+i)*m_strideRec[0] + curPosX];
				pRecLineBuf[i].U=m_pPelRec[1][((yCuStartInPic+i)>>uiScaleY)*m_strideRec[1] + (curPosX>>uiScaleX)];
				pRecLineBuf[i].V=m_pPelRec[2][((yCuStartInPic+i)>>uiScaleY)*m_strideRec[2] + (curPosX>>uiScaleX)];
			}
			else
			{
				pRecLineBuf[i].Y=m_pPelRec[0][(yCuStartInPic+i)*m_strideRec[0] + curPosX-1];
				pRecLineBuf[i].U=m_pPelRec[1][((yCuStartInPic+i)>>uiScaleY)*m_strideRec[1] + ((curPosX-1)>>uiScaleX)];
				pRecLineBuf[i].V=m_pPelRec[2][((yCuStartInPic+i)>>uiScaleY)*m_strideRec[2] + ((curPosX-1)>>uiScaleX)];
			}
		}
	}

	pixelCur = bUseOrg ? m_pOrgPixelBuffer[uiTraIdx] : m_pCQPixelBuffer[uiTraIdx];
	pixelAbv = bUseOrg ? m_pOrgPixelBuffer[uiTraIdxAbv] : m_pCQPixelBuffer[uiTraIdxAbv];  
	//search above
	if (abs(pixelCur.Y - pixelAbv.Y) < errorLimit && abs(pixelCur.U - pixelAbv.U) < errorLimit && abs(pixelCur.V - pixelAbv.V) < errorLimit )
	{
		pixelOrg = m_pOrgPixelBuffer[uiTraIdx];
		pixelRec = pRecLineBuf[(isHorizontalScanning ? offsetXInCu : offsetYInCu)];
		dist = CalcDist(pixelOrg, pixelRec);

		for (aboveLength = 1; cntNxt < uiTotal; aboveLength++)
		{
			uiIdxTrav = puiScanOrder[cntNxt];
			if(isHorizontalScanning)
			{
				offsetXInCu = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
				offsetYInCu = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;        
				curPosX = xCuStartInPic + offsetXInCu;
				curPosY = yCuStartInPic + offsetYInCu;
				//
				uiTraIdx= uiIdxTrav;
				uiTraIdxAbv = uiTraIdx - width;
			}
			else
			{ 
				offsetXInCu = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
				offsetYInCu = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;        
				curPosX = xCuStartInPic + offsetXInCu;
				curPosY = yCuStartInPic + offsetYInCu;
				//
				uiTraIdx= offsetXInCu + offsetYInCu * width;
				uiTraIdxAbv = uiTraIdx - 1;
			}

			pixelCur = bUseOrg ? m_pOrgPixelBuffer[uiTraIdx] : m_pCQPixelBuffer[uiTraIdx];
			pixelAbv = bUseOrg ? m_pOrgPixelBuffer[uiTraIdxAbv] : m_pCQPixelBuffer[uiTraIdxAbv];

			if (abs(pixelCur.Y - pixelAbv.Y) < errorLimit && abs(pixelCur.U - pixelAbv.U) < errorLimit && abs(pixelCur.V - pixelAbv.V) < errorLimit )
			{
				//
				pixelOrg = m_pOrgPixelBuffer[uiTraIdx];
				pixelRec = pRecLineBuf[(isHorizontalScanning ? offsetXInCu : offsetYInCu)];
				dist += CalcDist(pixelOrg, pixelRec);
				cntNxt++;
			}
			else
			{
				break;
			}
		}
	}

	if(pRecLineBuf)
	{
		delete [] pRecLineBuf;
		pRecLineBuf = NULL;
	}

	return aboveLength;
}

UInt TComPseudoDict::getSrbBits(TComDataCU*pcCU, TCombinePixel pixelCur, Int srbLength, UInt uiIdx, UInt uiTotal, UInt& uiSRBPos, UInt uiSRBPosPrev, ISC_TJ_MatchType MatchTypePrev)
{ 
	UInt srbBits;
	//Int  iMaxSymbol=m_uiSRBSize;
	Int  posInSRB=findPosInHiRefUmP( pixelCur, m_HiRefUmPixel, m_uiHiRefUmPSize);

	if(posInSRB == -1)
	{
		//new unmatched pixel
		uiSRBPos = m_uiHiRefUmPSize;
		srbBits  = EstWriteSRBPos(uiIdx, uiSRBPos, m_uiSRBSize, uiSRBPosPrev, MatchTypePrev);
		//ignore this
		//srb_bits +=Double(g_bitDepth[CHANNEL_TYPE_LUMA]  ) ;//m_pcBinIf->encodeBinsEP( Y, g_bitDepth[CHANNEL_TYPE_LUMA] );
		//srb_bits +=Double(g_bitDepth[CHANNEL_TYPE_CHROMA]) ;//m_pcBinIf->encodeBinsEP( U, g_bitDepth[CHANNEL_TYPE_CHROMA] );
		//srb_bits +=Double(g_bitDepth[CHANNEL_TYPE_CHROMA]) ;//m_pcBinIf->encodeBinsEP( V, g_bitDepth[CHANNEL_TYPE_CHROMA] );

		m_HiRefUmPixel[m_uiHiRefUmPSize] = pixelCur;
		m_uiHiRefUmPSize++;
	}
	else
	{
		uiSRBPos = posInSRB;
		srbBits  = EstWriteSRBPos(uiIdx, uiSRBPos, m_uiSRBSize, uiSRBPosPrev, MatchTypePrev);
	}
	srbBits += getLenBits(srbLength - 1, uiTotal - uiIdx-1  );//share the same length coding function
	//srbBits += getMTKLenBits(srbLength - 1, uiTotal - uiIdx-1  );
	//getLenBits(matchLength-1,cuWidth*cuWidth-1)
	return srbBits;
}

UInt TComPseudoDict::CalcDist( TCombinePixel orgPixel, TCombinePixel recPixel )
{
	UInt uiShiftY = DISTORTION_PRECISION_ADJUSTMENT((m_bitDepth[CHANNEL_TYPE_LUMA]-8) << 1);
	UInt uiShiftC = DISTORTION_PRECISION_ADJUSTMENT((m_bitDepth[CHANNEL_TYPE_CHROMA]-8) << 1);
	Int  iTemp;
	UInt uiSum=0;
#if	ISC_UV_DISTORTION_WEIGTH	
	iTemp = orgPixel.Y - recPixel.Y; uiSum += (Distortion)(lumaDistWeight*((( iTemp * iTemp ) >> uiShiftY)));
	iTemp = orgPixel.U - recPixel.U; uiSum += (Distortion)(chromaDistWeight*((( iTemp * iTemp ) >> uiShiftC)));
	iTemp = orgPixel.V - recPixel.V; uiSum += (Distortion)(chromaDistWeight*((( iTemp * iTemp ) >> uiShiftC)));
#else
	iTemp = orgPixel.Y - recPixel.Y; uiSum += (( iTemp * iTemp ) >> uiShiftY);
	iTemp = orgPixel.U - recPixel.U; uiSum += (( iTemp * iTemp ) >> uiShiftC);
	iTemp = orgPixel.V - recPixel.V; uiSum += (( iTemp * iTemp ) >> uiShiftC);
#endif
	return uiSum;
}

Int TComPseudoDict::EstWriteSRBPos(UInt uiIdx, UInt uiSRBPos, Int iMaxSymbol, UInt uiSRBPosPrev, ISC_TJ_MatchType MatchTypePrev)
{
	Int numBins=0;
	if (uiIdx && (MATCH_NONE != MatchTypePrev))
	{    
		if (MatchTypePrev == MATCH_POS_ONE)  //POS=1
		{
			assert(uiSRBPos != uiSRBPosPrev);
			if (uiSRBPos > uiSRBPosPrev)
			{
				uiSRBPos--;
			}
			iMaxSymbol--;
		}    
		/*else if (uiTraIdx >= iWidth && pMatchPoint[uiTraIdxLeft] == MATCH_POS_WIDTH && (pMatchPoint[uiTraIdxAbove] != MATCH_NONE))
		{
		assert(siCurLevel != pLevel[uiTraIdx - iWidth]);
		if (siCurLevel > pLevel[uiTraIdx - iWidth])
		{
		siCurLevel--;
		}
		iMaxSymbol--;
		}*/
	}
	assert(iMaxSymbol > 0);
	assert(uiSRBPos >= 0);
	assert(iMaxSymbol > uiSRBPos);
	if (iMaxSymbol > 1)
	{
		numBins = EstWriteTruncBinCode((UInt)uiSRBPos, iMaxSymbol);
	}
	return numBins;
}

Int TComPseudoDict::EstWriteTruncBinCode(UInt uiSymbol, UInt uiMaxSymbol)
{
	UInt uiThresh;
	if (uiMaxSymbol > 256)
	{
		UInt uiThreshVal = 1 << 8;
		uiThresh = 8;
		while (uiThreshVal <= uiMaxSymbol)
		{
			uiThresh++;
			uiThreshVal <<= 1;
		}
		uiThresh--;
	}
	else
	{
		uiThresh = g_uhPLTTBC[uiMaxSymbol];
	}

	UInt uiVal = 1 << uiThresh;
	assert(uiVal <= uiMaxSymbol);
	assert((uiVal << 1) > uiMaxSymbol);
	assert(uiSymbol < uiMaxSymbol);
	UInt b = uiMaxSymbol - uiVal;
	assert(b < uiVal);
	if (uiSymbol < uiVal - b)
	{
		return uiThresh;//m_pcBinIf->encodeBinsEP(uiSymbol, uiThresh);
	}
	else
	{
		uiSymbol += uiVal - b;
		assert(uiSymbol < (uiVal << 1));
		assert((uiSymbol >> 1) >= uiVal - b);
		return uiThresh + 1;//m_pcBinIf->encodeBinsEP(uiSymbol, uiThresh + 1);
	}
}

Bool TComPseudoDict::encode(TComDataCU* pcCU, Int width, Int height, Int xStartInPic, Int yStartInPic, vector<TDict1DInformation>& vecDictInfo, Double curBestRDCost)
{
	Int count = width*height;
	Int processedCount = 0;
#if STRING_COPY_RANGE != 1
	Int iLeftSR = pcCU->getStringMatchingSearchAreaWidth();
#endif
	const UInt  lcuWidth = pcCU->getSlice()->getSPS()->getMaxCUWidth();
	const UInt  lcuHeight = pcCU->getSlice()->getSPS()->getMaxCUHeight();
	UInt uiScaleX = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);
	UInt uiScaleY = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
	//ChromaFormat chFormat = pcCU->getPic()->getChromaFormat();
	//Bool isprintf = false;
	Bool isHorizontalScanning = pcCU->getDictStatus(0) == DICT_STATUS_REC_HORIZONTAL ? true : false;
	Int  CbSize    = width;
	Int  log2CbSize=g_aucConvertToBit[width]  + 2;
	Int  TravX;
	Int  TravY;	
	Int  errorLimit = pcCU->getCUTransquantBypass(0) ? 1 : 1 + g_uhPLTQuant[Int(pcCU->getQP(0))];
	Double DictRDCost=0;
	Double tmpDictRDCost=0;
	UInt* puiScanOrder = g_scanOrder[SCAN_UNGROUPED][SCAN_TRAV][g_aucConvertToBit[width] + 2][g_aucConvertToBit[height] + 2];
	UInt uiIdxTrav;
#if STRING_COPY_RANGE == 1
	Int ctuSizeLog2 = g_aucConvertToBit[pcCU->getSlice()->getSPS()->getMaxCUWidth()] + 2;
#endif
#if ISC_STR_LEN_MIN
	Int lenLimitLeftWt = pcCU->getStringLengthLimitLeftWidth();
	Int lcuStartXInPic = pcCU->getPic()->getCtu(pcCU->getCtuRsAddr())->getCUPelX();
	Int lcuStartYInPic = pcCU->getPic()->getCtu(pcCU->getCtuRsAddr())->getCUPelY();
#endif
	Bool bLossless = pcCU->getCUTransquantBypass (0);
	Int  srbLength;
	Int  srbAboveLength;
	UInt srbAboveDist;
	UInt srbAboveBits=0;
	Double srbAboveRdcost;
	ISC_TJ_MatchType srbMatchMode;
	UInt uiSRBPos=0;
	UInt uiSRBPosPrev=0;
	ISC_TJ_MatchType MatchTypePrev = MATCH_NONE;
	m_bHiRefFlag=false;
	m_bLoRefFlag=false; 
	UInt uiMaxVal[3];
  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
  {
    uiMaxVal[ch] = pcCU->xCalcMaxVals(pcCU, ComponentID(ch));
  }

	UInt uiDepth = pcCU->getDepth(0);
	Int iCUZOrderStart = pcCU->getZorderIdxInCtu();
	Int uiEncodedPixelCountInLcu = (iCUZOrderStart << 4) ;
	Short** pZOrder;
	if(isHorizontalScanning)
		pZOrder = g_pixOrderH[uiDepth];
	else
		pZOrder = g_pixOrderV[uiDepth];
	UInt uiShiftY = DISTORTION_PRECISION_ADJUSTMENT((m_bitDepth[CHANNEL_TYPE_LUMA]-8) << 1);
	UInt uiShiftC = DISTORTION_PRECISION_ADJUSTMENT((m_bitDepth[CHANNEL_TYPE_CHROMA]-8) << 1);
	Int  iTemp=0;
	UInt uiSum=0;

  while ( processedCount < count )
		{
			Bool bFound = false;
			const Int maxTestCases = MAX_DICT_TEST_CASE;
			Int attemp = 0;
			Int currentX;// = xStartInPic + processedCount%width;
			Int currentY;// = yStartInPic + processedCount/width;
			uiIdxTrav = puiScanOrder[processedCount];
			TravX=getTravX(uiIdxTrav, CbSize    ); //assert(TravX==(uiIdxTrav%width));
			TravY=getTravY(uiIdxTrav, log2CbSize); //assert(TravY==(uiIdxTrav/width));
			if ( isHorizontalScanning )
			{
				currentX = xStartInPic + TravX;
				currentY = yStartInPic + TravY;
			}
			else
			{
				currentX = xStartInPic + TravY;
				currentY = yStartInPic + TravX;
				uiIdxTrav= TravX * width + TravY;
			}
			Int curOffset = currentY * m_strideOrg[0] + currentX;
			Int curOffsetC= (currentY>>uiScaleY) * m_strideOrg[1] + (currentX>>uiScaleX);
			Int maxMatchLength = 0;
			Int bestMatchStartX = 0;
			Int bestMatchStartY = 0;
			UInt curDistortion = 0;
			Double curRdCost = 0;
			Double minRdCost = MAX_DOUBLE;  
#if ISC_STR_LEN_MIN
			Int maxMatchLengthInCU = 0;
			Int bestMatchStartXInCU = 0;
			Int bestMatchStartYInCU = 0;
			double bestRdCostInCU = 0;
#endif

      //do srb search
			TCombinePixel pixelCur;
			Double prb_rdcost=0.0, srb_rdcost=0.0;
			UInt  srb_dist=0;
			UInt  srb_bits=0;
			UInt  curSRBSize = m_uiHiRefUmPSize;

			Int prb_length  = 0;

			srbLength = getSrbLength(pcCU, width, height, xStartInPic, yStartInPic, processedCount, srb_dist, pixelCur);
			if(srbLength>0)
				srb_bits = getSrbBits(pcCU, pixelCur, srbLength, processedCount, count, uiSRBPos, uiSRBPosPrev, MatchTypePrev);
			else
			{
				srb_bits = EstWriteSRBPos(processedCount, m_uiSRBSize, m_uiSRBSize+1, uiSRBPosPrev, MatchTypePrev);
				srb_bits += 24;
			}
			srb_rdcost = m_pRdCostImp->calcRdCost(srb_bits,srb_dist);

			srbAboveLength = getSrbAboveLength(pcCU, width, height, xStartInPic, yStartInPic, processedCount, srbAboveDist, pixelCur);
			if(srbAboveLength>0)
			{
				srbAboveBits = 2;
				srbAboveBits += getLenBits(srbAboveLength - 1, count - processedCount - 1 );//share the same length coding function		 
				srbAboveRdcost=m_pRdCostImp->calcRdCost(srbAboveBits,srbAboveDist);
			}
			else
				srbAboveRdcost=MAX_DOUBLE;
			srbMatchMode = MATCH_POS_ONE;
			if( srbAboveLength > 0 )
			{
				if( srbLength > 0 )
				{
					if(srbAboveLength>=srbLength && srbAboveRdcost * srbLength <= srb_rdcost * srbAboveLength)
						srbMatchMode = MATCH_POS_WIDTH;
				}
				else
					srbMatchMode = MATCH_POS_WIDTH;

				if(srbMatchMode == MATCH_POS_WIDTH)
				{
					srbLength = srbAboveLength;
					srb_bits  = srbAboveBits;
					srb_dist  = srbAboveDist;
					srb_rdcost= srbAboveRdcost;
				}
			}
			if(srbLength>0)
			{
				UInt maxLenTemp = (srbLength+3)>( count - processedCount )?( count - processedCount):(srbLength+3);
				maxMatchLength = maxLenTemp;
			}
			//maxMatchLength=0;
			tmpDictRDCost = srb_rdcost; //save current best srb RDCost

			{			
				Int currMatchEndOffsetX = 0;
				Int currMatchEndOffsetY = 0;
				UShort hashValue = (this->*getCRCValue)(m_pOrgPixelBuffer[uiIdxTrav]);
				Int currMatchStartX = m_hashTable[hashValue].x;
				Int currMatchStartY = m_hashTable[hashValue].y;
				for ( ; attemp < maxTestCases && currMatchStartX >= 0 && currMatchStartY >= 0; attemp++ )
				{
					Int nextMatchStartX = m_dict[currMatchStartY + currMatchEndOffsetY][currMatchStartX + currMatchEndOffsetX].x;
					Int nextMatchStartY = m_dict[currMatchStartY + currMatchEndOffsetY][currMatchStartX + currMatchEndOffsetX].y;
					Int currMatchLength = 0;
					//Bool bCheckCondition= true;
					Int refOffset = 0;
					Int refOffsetC= 0;
					//{
					Int mvX = currMatchStartX - currentX;
					Int mvY = currMatchStartY - currentY;
					Int iRefX = (currentX & (lcuWidth-1)) + mvX;//((xCuStartInPic + iCurX) % lcuWidth) + mvX;
					Int iRefY = (currentY & (lcuHeight-1)) + mvY;//((yCuStartInPic + iCurY) % lcuHeight) + mvY; 
#if STRING_COPY_RANGE != 1
					//Test if the reference pixel is within Search Range
					if ( iRefX < -iLeftSR || iRefX >= (Int)lcuWidth || iRefY >= (Int)lcuHeight || iRefY < 0 ) 
					{

						//bCheckCondition=false; //out of the valid research range
						break;
					}
#endif
					//	}
					//	if(bCheckCondition)
					//	{
					refOffset = currMatchStartY * m_strideOrg[0] + currMatchStartX;
					refOffsetC= (currMatchStartY>>uiScaleY) * m_strideOrg[1] + (currMatchStartX>>uiScaleX);
					//if (abs(m_pPelOrg[0][curOffset] - m_pPelOrg[0][refOffset]) < errorLimit && abs(m_pPelOrg[1][curOffset] - m_pPelOrg[1][refOffset]) < errorLimit && abs(m_pPelOrg[2][curOffset] - m_pPelOrg[2][refOffset]) < errorLimit)
					if (abs(m_pPelOrg[0][curOffset] - m_pPelOrg[0][refOffset]) >= errorLimit)
					{
						currMatchStartX = nextMatchStartX;
						currMatchStartY = nextMatchStartY;
						continue;
					}

					//	}
					Int uiIdxMaxLenMinus1Travt=0;
					Int iCurXMLMinus1,iCurYMLMinus1;
					Int iRefXInPicN ;
					Int iRefYInPicN ;
					if(maxMatchLength>2)
					{
						uiIdxMaxLenMinus1Travt = puiScanOrder[processedCount+maxMatchLength-2];	
						if ( isHorizontalScanning )
						{
							iCurXMLMinus1 = getTravX(uiIdxMaxLenMinus1Travt, CbSize    );
							iCurYMLMinus1 = getTravY(uiIdxMaxLenMinus1Travt, log2CbSize);
						}
						else
						{
							iCurXMLMinus1 = getTravY(uiIdxMaxLenMinus1Travt, log2CbSize);
							iCurYMLMinus1 = getTravX(uiIdxMaxLenMinus1Travt, CbSize    );
						}							
						iRefXInPicN = xStartInPic+iCurXMLMinus1 + mvX;
						iRefYInPicN = yStartInPic+iCurYMLMinus1 + mvY;
						Int curNOffset = (yStartInPic + iCurYMLMinus1) * m_strideOrg[0] + xStartInPic + iCurXMLMinus1;
						Int refNOffset = iRefYInPicN * m_strideOrg[0] + iRefXInPicN;		
						if (abs(m_pPelOrg[0][curNOffset] - m_pPelOrg[0][refNOffset]) >= errorLimit)
						{
							currMatchStartX = nextMatchStartX;
							currMatchStartY = nextMatchStartY;
							continue;
						}
						Int iRefXN = ((xStartInPic+iCurXMLMinus1) & (lcuWidth-1)) + mvX;//((xCuStartInPic + iCurX) % lcuWidth) + mvX;
						Int iRefYN = ((yStartInPic+iCurYMLMinus1) & (lcuHeight-1)) + mvY;//((yCuStartInPic + iCurY) % lcuHeight) + mvY;
#if STRING_COPY_RANGE == 1					
						if ((iRefYInPicN >> ctuSizeLog2) > (yStartInPic >> ctuSizeLog2))
						{
							currMatchStartX = nextMatchStartX;
							currMatchStartY = nextMatchStartY;
							continue;
						}
						else if ((iRefYInPicN >> ctuSizeLog2) == (yStartInPic >> ctuSizeLog2))
						{
							if ((iRefXInPicN >> ctuSizeLog2) > (xStartInPic>> ctuSizeLog2))
							{
								currMatchStartX = nextMatchStartX;
								currMatchStartY = nextMatchStartY;
								continue;
							}
						}
#if ISC_WPP_STEP
						if (!isValidIntraSCSearchArea(pcCU, mvX, mvY, 0))
						{
							currMatchStartX = nextMatchStartX;
							currMatchStartY = nextMatchStartY;
							continue;
						}
#endif
#else	
						//Test if the reference pixel is within Search Range
						if ( iRefXN < -iLeftSR || iRefXInPicN >= m_picWidth || iRefXN >= (Int)lcuWidth || iRefYInPicN >= m_picHeight || iRefYN >= (Int)lcuHeight || iRefYN < 0 )
						{
							currMatchStartX = nextMatchStartX;
							currMatchStartY = nextMatchStartY;
							continue;
						}
#endif
						if ((iRefXN >= 0) && (iRefYN >= 0)) //in current LCU
						{
							UInt uiepc=uiEncodedPixelCountInLcu+(processedCount+maxMatchLength-2);
							if(pZOrder[iRefYN][iRefXN]>uiepc)	
							{
								currMatchStartX = nextMatchStartX;
								currMatchStartY = nextMatchStartY;
								continue;
							}
						}				
					}

          //HASH search
          {
            UInt pc = processedCount;
            UInt cMatchLength = 0;
            //Int curPos = curOffset; 
            Int refPos = refOffset;
            Int refPosC= refOffsetC;
            UInt uEpcInLcu =uiEncodedPixelCountInLcu+pc;
            uiIdxTrav = puiScanOrder[processedCount];
            if ( !isHorizontalScanning )
            {
              TravX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
              TravY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
              uiIdxTrav=TravY*CbSize + TravX;
            }
            if (abs(m_pOrgPixelBuffer[uiIdxTrav].Y - m_pPelOrg[0][refPos ]) < errorLimit && 
                abs(m_pOrgPixelBuffer[uiIdxTrav].U - m_pPelOrg[1][refPosC]) < errorLimit && 
                abs(m_pOrgPixelBuffer[uiIdxTrav].V - m_pPelOrg[2][refPosC]) < errorLimit)
            {
              cMatchLength++;
              pc++;	
              while (pc < count)
              {
                Int iCurX, iCurY;
                uiIdxTrav = puiScanOrder[pc];
                if ( isHorizontalScanning )
                {
                  iCurX = getTravX(uiIdxTrav, CbSize    );
                  iCurY = getTravY(uiIdxTrav, log2CbSize);
                }
                else
                {
                  iCurX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
                  iCurY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
                  uiIdxTrav=iCurY*CbSize + iCurX;
                }
                iRefX = ((xStartInPic + iCurX) & (lcuWidth-1)) + mvX;//((xCuStartInPic + iCurX) % lcuWidth) + mvX;
                iRefY = ((yStartInPic + iCurY) & (lcuHeight-1)) + mvY;//((yCuStartInPic + iCurY) % lcuHeight) + mvY;
                Int iRefXInPic = xStartInPic + iCurX + mvX;
                Int iRefYInPic = yStartInPic + iCurY + mvY;
#if STRING_COPY_RANGE == 1
                if (iRefYInPic < 0 || iRefYInPic >= m_picHeight)
                  break;
                if (iRefXInPic < 0 || iRefXInPic >= m_picWidth)
                  break;
                if ((iRefYInPic >> ctuSizeLog2) > (yStartInPic >> ctuSizeLog2))
                {
                  break;
                }
                else if ((iRefYInPic >> ctuSizeLog2) == (yStartInPic >> ctuSizeLog2))
                {
                  if ((iRefXInPic >> ctuSizeLog2) > (xStartInPic >> ctuSizeLog2))
                    break;
                }
#if ISC_WPP_STEP
                if (!isValidIntraSCSearchArea(pcCU, mvX, mvY, 0))
                  break;
#endif
#else
                //Test if the reference pixel is within Search Range
                if ( iRefX < -iLeftSR || iRefXInPic >= m_picWidth || iRefX >= (Int)lcuWidth || iRefYInPic >= m_picHeight || iRefY >= (Int)lcuHeight || iRefY < 0 )
                {
                  break; //out of the valid research range
                }
#endif	
                if ((iRefX >= 0) && (iRefY >= 0)) //in current LCU
                {

                  if(pZOrder[iRefY][iRefX]>uEpcInLcu)	
                  {

                    break;	
                  }
                }
                //Int offset = travDateOrder[pc-1];
                //curPos +=offset;
                //refPos +=offset;
                refPos = iRefYInPic * m_strideOrg[0] + iRefXInPic;
                refPosC= (iRefYInPic>>uiScaleY) * m_strideOrg[1] + (iRefXInPic>>uiScaleX);
                if (abs(m_pOrgPixelBuffer[uiIdxTrav].Y - m_pPelOrg[0][refPos]) < errorLimit && 
                    abs(m_pOrgPixelBuffer[uiIdxTrav].U - m_pPelOrg[1][refPosC]) < errorLimit && 
                    abs(m_pOrgPixelBuffer[uiIdxTrav].V - m_pPelOrg[2][refPosC]) < errorLimit)
                {
                  cMatchLength++;
                  pc++;
                  uEpcInLcu++;
                }
                else
                  break;

              }
            }
            currMatchLength = cMatchLength;
          }

					//if(bCheckCondition)
					{
						Bool isBCU = false;
						isBCU = pcCU->isBoudaryCU(pcCU->getZorderIdxInCtu(),pcCU->getDepth(0),pcCU->getDictStatus(0));
						curDistortion = 0;
						if ( isHorizontalScanning )
						{
							UInt offsetInCU=0;
							Int mxLen = (maxMatchLength==0)?0:maxMatchLength-1;
							if(currMatchLength!=0  &&currMatchLength>=mxLen)
							{
								uiIdxTrav = puiScanOrder[processedCount];
								UInt offsetSignFlag  = 0;
								UInt offsetXInCU = getTravX(uiIdxTrav, CbSize    );
								UInt offsetYInCU = getTravY(uiIdxTrav, log2CbSize);
								UInt offsetMapValue = offsetXInCU;
								if( 1 == (offsetYInCU & 0x1) )
								{				
									offsetSignFlag  = 1;
								}
								if( 1 == (offsetYInCU & 0x1))
									offsetMapValue++;
								else
									offsetMapValue += currMatchLength;			
								offsetInCU = offsetMapValue>=width?width:offsetMapValue;
								offsetInCU--;

							if(errorLimit!=1)
							{
								Int curPos = curOffset;
								Int refPos = refOffset;
								Int curPosC= curOffsetC;
								Int refPosC= refOffsetC;
								for (UInt pixelCount = 0; pixelCount < currMatchLength; pixelCount++)
								{
									uiIdxTrav = puiScanOrder[processedCount + pixelCount];
									if ( isHorizontalScanning )
									{
										TravX = getTravX(uiIdxTrav, CbSize    );
										TravY = getTravY(uiIdxTrav, log2CbSize);
									}
									else
									{
										TravX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
										TravY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
										uiIdxTrav=TravY*CbSize + TravX;
									}
									Int iCurXInPic = xStartInPic + TravX;
									Int iCurYInPic = yStartInPic + TravY;
									Int iRefXInPic = xStartInPic + TravX + mvX;
									Int iRefYInPic = yStartInPic + TravY + mvY;
									curPos = iCurYInPic * m_strideRec[0] + iCurXInPic;
									curPosC= (iCurYInPic>>uiScaleY) * m_strideRec[1] + (iCurXInPic>>uiScaleX);
									refPos = iRefYInPic * m_strideRec[0] + iRefXInPic;
									refPosC= (iRefYInPic>>uiScaleY) * m_strideRec[1] + (iRefXInPic>>uiScaleX);
									uiSum = 0;
#if	ISC_UV_DISTORTION_WEIGTH
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].Y- m_pPelRec[0][refPos ]; uiSum += (Distortion)(lumaDistWeight*((( iTemp * iTemp ) >> uiShiftY)));
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].U- m_pPelRec[1][refPosC]; uiSum += (Distortion)(chromaDistWeight*((( iTemp * iTemp ) >> uiShiftC)));
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].V- m_pPelRec[2][refPosC]; uiSum += (Distortion)(chromaDistWeight*((( iTemp * iTemp ) >> uiShiftC)));
#else
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].Y- m_pPelRec[0][refPos]; uiSum += (( iTemp * iTemp ) >> uiShiftY);
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].U- m_pPelRec[1][refPosC]; uiSum += (( iTemp * iTemp ) >> uiShiftC);
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].V- m_pPelRec[2][refPosC]; uiSum += (( iTemp * iTemp ) >> uiShiftC);
#endif
									curDistortion+=uiSum;
									m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
									//m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
									//m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
									if(!((iCurXInPic & uiScaleX) || (iCurYInPic & uiScaleY)))
									{
										m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
										m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
									}
								}
							}
							curRdCost = getDictCost(currentX - currMatchStartX,currentY - currMatchStartY,curDistortion,offsetInCU,isBCU,width,currMatchLength,offsetSignFlag,count-processedCount-1);

							}
							else
								curRdCost = MAX_DOUBLE;				
						}
						else
						{
							UInt offsetInCU=0;
							Int mxLen = (maxMatchLength==0)?0:maxMatchLength-1;
							if(currMatchLength!=0  &&currMatchLength>=mxLen)
							{
								uiIdxTrav = puiScanOrder[processedCount];
								UInt offsetSignFlag  = 0;
								UInt offsetXInCU = getTravY(uiIdxTrav, log2CbSize);
								UInt offsetYInCU = getTravX(uiIdxTrav, CbSize    );
								UInt offsetMapValue = offsetYInCU;

								if( 1 == (offsetXInCU & 0x1) )
								{			
									offsetSignFlag  = 1;
								}
								if( 1 == (offsetXInCU & 0x1) )
									offsetMapValue++;
								else
									offsetMapValue += currMatchLength;
								offsetInCU = offsetMapValue>=height?height:offsetMapValue;
								offsetInCU--;
							if(errorLimit!=1)
							{
								Int curPos = curOffset;
								Int refPos = refOffset;
								Int curPosC= curOffsetC;
								Int refPosC= refOffsetC;
								uiSum = 0;
								for (UInt pixelCount = 0; pixelCount < currMatchLength; pixelCount++)
								{
									uiIdxTrav = puiScanOrder[processedCount + pixelCount];
									if ( isHorizontalScanning )
									{
										TravX = getTravX(uiIdxTrav, CbSize    );
										TravY = getTravY(uiIdxTrav, log2CbSize);
									}
									else
									{
										TravX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
										TravY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
										uiIdxTrav=TravY*CbSize + TravX;
									}
									Int iCurXInPic = xStartInPic + TravX;
									Int iCurYInPic = yStartInPic + TravY;
									Int iRefXInPic = xStartInPic + TravX + mvX;
									Int iRefYInPic = yStartInPic + TravY + mvY;
									curPos = iCurYInPic * m_strideRec[0] + iCurXInPic;
									curPosC= (iCurYInPic>>uiScaleY) * m_strideRec[1] + (iCurXInPic>>uiScaleX);
									refPos = iRefYInPic * m_strideRec[0] + iRefXInPic;                  
									refPosC= (iRefYInPic>>uiScaleY) * m_strideRec[1] + (iRefXInPic>>uiScaleX);
									uiSum = 0;
#if	ISC_UV_DISTORTION_WEIGTH
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].Y- m_pPelRec[0][refPos ]; uiSum += (Distortion)(lumaDistWeight*((( iTemp * iTemp ) >> uiShiftY)));
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].U- m_pPelRec[1][refPosC]; uiSum += (Distortion)(chromaDistWeight*((( iTemp * iTemp ) >> uiShiftC)));
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].V- m_pPelRec[2][refPosC]; uiSum += (Distortion)(chromaDistWeight*((( iTemp * iTemp ) >> uiShiftC)));
#else
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].Y- m_pPelRec[0][refPos]; uiSum += (( iTemp * iTemp ) >> uiShiftY);
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].U- m_pPelRec[1][refPosC]; uiSum += (( iTemp * iTemp ) >> uiShiftC);
									iTemp = m_pOrgPixelBuffer[uiIdxTrav].V- m_pPelRec[2][refPosC]; uiSum += (( iTemp * iTemp ) >> uiShiftC);
#endif
									curDistortion+=uiSum;
									m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
									//m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
									//m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
									if(!((iCurXInPic & uiScaleX) || (iCurYInPic & uiScaleY)) )
									{
										m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
										m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
									}
								}
							}
								curRdCost = getVerDictCost(currentX - currMatchStartX,currentY - currMatchStartY,curDistortion,offsetInCU,isBCU,height,currMatchLength,offsetSignFlag,count-processedCount-1);	
							}
							else
								curRdCost = MAX_DOUBLE;
						}
						//if(curRdCost!=MAX_DOUBLE) printf("Hash:%1d %4d %4d %4d %8d %8d\n",isHorizontalScanning,processedCount,attemp,currMatchLength,curDistortion,Int(curRdCost));
					}//enc checkCondition
					if ( currMatchLength > maxMatchLength 
						||((currMatchLength == maxMatchLength)  && currMatchLength!=0 && curRdCost<minRdCost )  
						||(bLossless && (currMatchLength == maxMatchLength-1)  && currMatchLength!=0 && (curRdCost*128)<minRdCost )  
						||(!bLossless && (currMatchLength == maxMatchLength-1)  && currMatchLength!=0 && (curRdCost*maxMatchLength)<(minRdCost*currMatchLength) )  
						)//ZF RD encoder need to modify
					{
#if ISC_STR_LEN_MIN
						if((currMatchStartX >= lcuStartXInPic- lenLimitLeftWt) && currMatchStartY>=lcuStartYInPic )
						{
							maxMatchLengthInCU =  currMatchLength;
							bestMatchStartXInCU = currMatchStartX;
							bestMatchStartYInCU = currMatchStartY;
							bestRdCostInCU = curRdCost;

						}
#endif
						bFound = true;
						maxMatchLength = currMatchLength;
						bestMatchStartX = currMatchStartX;
						bestMatchStartY = currMatchStartY;
						minRdCost = curRdCost;
					}

					currMatchStartX = nextMatchStartX;
					currMatchStartY = nextMatchStartY;
					if ( processedCount+currMatchLength == count )
					{
						break;
					}
				}
#if ISC_STR_LEN_MIN
				if ( bFound )
				{
					if(maxMatchLength< ISC_STR_LEN_MIN)
					{
						//assert(maxMatchLength>maxMatchLengthInCU);
						if(maxMatchLengthInCU == 0)
							bFound = false;
						else 
						{
							maxMatchLength = maxMatchLengthInCU;
							bestMatchStartX = bestMatchStartXInCU;
							bestMatchStartY = bestMatchStartYInCU;
							minRdCost = bestRdCostInCU;
						}
					}
				}
#endif
			} // end hash search

			{
				prb_rdcost = minRdCost;
				prb_length = maxMatchLength;
				if ( bFound && srbLength>0)
				{
					if(srbLength >= maxMatchLength && srb_rdcost * prb_length <= prb_rdcost * srbLength)
						bFound = false;
					else //if(0)
					{
						uiSRBPosPrev = uiSRBPos;
						MatchTypePrev= srbMatchMode;//MATCH_POS_ONE;
						//if(srb_rdcost * prb_length < prb_rdcost * srbLength && (srbLength * 4) > (prb_length * 3))
						//bFound = false;
						if(srb_rdcost < prb_rdcost && (prb_length + processedCount) < count)
						{
							Double NxtSrbRdCost;
							UInt  NxtSrbDist;
							UInt  NxtSrbBits;
							Int   NxtSrbLength=0;
							//
							UInt  NxtsrbAboveDist = 0;
							Int   NxtsrbAboveLength = 0;
							UInt  NxtsrbAboveBits = 0;
							Double NxtsrbAboveRdcost = 0.0;
							//
							Int TotalSrbLength=srbLength;
							Int startPos = processedCount + srbLength, searchLen = prb_length - srbLength;
							Int loopCnt = 0;
							while((searchLen > 0) && (srb_rdcost < prb_rdcost) && loopCnt<2)
							{
								NxtSrbLength = getSrbLength(pcCU, width, height, xStartInPic, yStartInPic, startPos, NxtSrbDist, pixelCur);
								if(NxtSrbLength>0)
									NxtSrbBits = 1+getSrbBits(pcCU, pixelCur, NxtSrbLength, startPos, count, uiSRBPos, uiSRBPosPrev, MatchTypePrev);
								else
								{
									NxtSrbBits = EstWriteSRBPos(startPos, m_uiSRBSize, m_uiSRBSize+1, uiSRBPosPrev, MatchTypePrev);
									NxtSrbBits += 24;
								}
								NxtSrbRdCost = m_pRdCostImp->calcRdCost(NxtSrbBits,NxtSrbDist);               

								if(NxtSrbLength > 0)
								{
									uiSRBPosPrev = uiSRBPos;
									MatchTypePrev=MATCH_POS_ONE;
									NxtsrbAboveLength = getSrbAboveLength(pcCU, width, height, xStartInPic, yStartInPic, startPos, NxtsrbAboveDist, pixelCur);
									if(NxtsrbAboveLength>0)
									{
										NxtsrbAboveBits = 2;
										NxtsrbAboveBits += getLenBits(NxtsrbAboveLength - 1, count - startPos - 1 );//share the same length coding function		 
										NxtsrbAboveRdcost=m_pRdCostImp->calcRdCost(NxtsrbAboveBits,NxtsrbAboveDist);
									}
									else
										NxtsrbAboveRdcost=MAX_DOUBLE;
									if( NxtsrbAboveLength> 0 )
									{
										if( NxtSrbLength > 0 )
										{
											if(NxtsrbAboveLength>=NxtSrbLength && NxtsrbAboveRdcost * NxtSrbLength <= NxtSrbRdCost * NxtsrbAboveLength)
												MatchTypePrev = MATCH_POS_WIDTH;            
										}
										else
											MatchTypePrev = MATCH_POS_WIDTH;

										if(MatchTypePrev == MATCH_POS_WIDTH)
										{
											NxtSrbLength = NxtsrbAboveLength;
											NxtSrbBits  = NxtsrbAboveBits;
											NxtSrbDist  = NxtsrbAboveDist;
										}
									}  

									srb_bits += NxtSrbBits;
									srb_dist += NxtSrbDist;
									srb_rdcost = m_pRdCostImp->calcRdCost(srb_bits,srb_dist);
									loopCnt++;
									TotalSrbLength += NxtSrbLength;
									startPos += NxtSrbLength;
									searchLen -= NxtSrbLength;
								}
								else
									break;
							}
							//if(srb_rdcost * prb_length < prb_rdcost * TotalSrbLength && (TotalSrbLength * 8) > (prb_length * 7))
							if(srb_rdcost * prb_length < prb_rdcost * TotalSrbLength)
								bFound = false;
						}
					}
				}
				//
				uiSRBPosPrev=0; MatchTypePrev=MATCH_NONE;
				m_uiHiRefUmPSize = curSRBSize;
			}

			if ( bFound )
			{
				tmpDictRDCost = minRdCost; //choose Hash RDCost
				TDict1DInformation dictInfo;
				dictInfo.isMatched = true;
				dictInfo.offsetX = currentX - bestMatchStartX;
				dictInfo.offsetY = currentY - bestMatchStartY;
				dictInfo.length = maxMatchLength;				
				//reconstruction
				for (UInt pixelCount = 0; pixelCount < dictInfo.length; pixelCount++)
				{
					Int curPosX;
					Int curPosY;
					Int curPixelPos = puiScanOrder[processedCount + pixelCount];
					TravX=getTravX(curPixelPos, CbSize    );
					TravY=getTravY(curPixelPos, log2CbSize);
					if( isHorizontalScanning )
					{
						curPosX = xStartInPic + TravX;//curPixelPos % width;
						curPosY = yStartInPic + TravY;//curPixelPos / width;
					}
					else
					{
						curPosX = xStartInPic + TravY;//curPixelPos / height;
						curPosY = yStartInPic + TravX;//curPixelPos % height;
					}
					Int curPos = curPosY * m_strideRec[0] + curPosX;
					Int curPosC= (curPosY>>uiScaleY) * m_strideRec[1] + (curPosX>>uiScaleX);
					Int refPos = (curPosY - dictInfo.offsetY) * m_strideRec[0] + (curPosX - dictInfo.offsetX);
					Int refPosC= ((curPosY- dictInfo.offsetY)>>uiScaleY) * m_strideRec[1] + ((curPosX- dictInfo.offsetX)>>uiScaleX);

					m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
					//m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
					//m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
					if(!((curPosX & uiScaleX) || (curPosY & uiScaleY)) )
					{
						m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
						m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
					}
				}

				dictInfo.MatchType = MATCH_NONE; uiSRBPosPrev=0; MatchTypePrev=MATCH_NONE;
				//add to hash 
				addPixelToDict(maxMatchLength, true, xStartInPic, yStartInPic, width, height, processedCount, isHorizontalScanning);
				vecDictInfo.push_back( dictInfo );
				processedCount += maxMatchLength;
			}
			else
			{
				TDict1DInformation dictInfo;
				dictInfo.isMatched = false;
				dictInfo.pixel.Y = m_pPelOrg[0][currentY*m_strideOrg[0] + currentX];
				dictInfo.pixel.U = m_pPelOrg[1][(currentY>>uiScaleY)*m_strideOrg[1] + (currentX>>uiScaleX)];
				dictInfo.pixel.V = m_pPelOrg[2][(currentY>>uiScaleY)*m_strideOrg[2] + (currentX>>uiScaleX)];
				dictInfo.length = srbLength;
				if(dictInfo.length == -1)
					dictInfo.length=1;

				if(srbMatchMode == MATCH_POS_WIDTH)
				{ 			
					for (UInt pixelCount = 0; pixelCount < dictInfo.length; pixelCount++)
					{
						Int curPosX;
						Int curPosY;
						Int curPos;
						Int curPosC;
						Int refPos ;
						Int refPosC;
						Int curPixelPos = puiScanOrder[processedCount + pixelCount];
						TravX=getTravX(curPixelPos, CbSize    );
						TravY=getTravY(curPixelPos, log2CbSize);
						if( isHorizontalScanning )
						{
							curPosX = xStartInPic + TravX;//curPixelPos % width;
							curPosY = yStartInPic + TravY;//curPixelPos / width;
							refPos = (curPosY-1) * m_strideRec[0] + curPosX;
							refPosC= ((curPosY-1)>>uiScaleY) * m_strideRec[1] + (curPosX>>uiScaleX);							
						}
						else
						{
							curPosX = xStartInPic + TravY;//curPixelPos / height;
							curPosY = yStartInPic + TravX;//curPixelPos % height;
							refPos = curPosY * m_strideRec[0] + (curPosX-1);
							refPosC= (curPosY>>uiScaleY) * m_strideRec[1] + ((curPosX-1)>>uiScaleX);
						}
						curPos = curPosY * m_strideRec[0] + curPosX;
						curPosC= (curPosY>>uiScaleY) * m_strideRec[1] + (curPosX>>uiScaleX);
						m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
						//m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
						//m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
						if(!((curPosX & uiScaleX) || (curPosY & uiScaleY)) )
						{
							m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
							m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
						}
					}
					dictInfo.MatchType = MATCH_POS_WIDTH; uiSRBPosPrev=0; MatchTypePrev=MATCH_POS_WIDTH;
				}
				else
				{
					if(srbLength > 0) //if(m_bMajor[uiTraIdx] == true)
					{
						m_bHiRefFlag = true;
						UInt uiTraIdx = (currentX - xStartInPic) + (currentY - yStartInPic) * width;
						dictInfo.pixel  = m_pCQPixelBuffer[uiTraIdx];
						Int posInSRB=findPosInHiRefUmP( dictInfo.pixel, m_HiRefUmPixel, m_uiHiRefUmPSize);
						if(posInSRB == -1)
						{
							//new unmatched pixel
							m_HiRefUmPixel[m_uiHiRefUmPSize] = dictInfo.pixel;
							dictInfo.srbIndex = m_uiHiRefUmPSize;
							m_uiHiRefUmPSize++;
							assert( m_uiHiRefUmPSize <=m_MaxPLTSize);
						}
						else
							dictInfo.srbIndex = posInSRB;
						// reconstruction
						for (UInt pixelCount = 0; pixelCount < dictInfo.length; pixelCount++)
						{
							Int curPosX;
							Int curPosY;
							Int curPixelPos = puiScanOrder[processedCount + pixelCount];
							TravX=getTravX(curPixelPos, CbSize    );
							TravY=getTravY(curPixelPos, log2CbSize);
							if( isHorizontalScanning )
							{
								curPosX = xStartInPic + TravX;//curPixelPos % width;
								curPosY = yStartInPic + TravY;//curPixelPos / width;
							}
							else
							{
								curPosX = xStartInPic + TravY;//curPixelPos / height;
								curPosY = yStartInPic + TravX;//curPixelPos % height;
							}
							Int curPos = curPosY * m_strideRec[0] + curPosX;
							Int curPosC= (curPosY>>uiScaleY) * m_strideRec[1] + (curPosX>>uiScaleX);
							m_pPelRec[0][curPos ] = dictInfo.pixel.Y;
							//m_pPelRec[1][curPosC] = dictInfo.pixel.U;
							//m_pPelRec[2][curPosC] = dictInfo.pixel.V;
							if(!((curPosX & uiScaleX) || (curPosY & uiScaleY)) )
							{
								m_pPelRec[1][curPosC] = dictInfo.pixel.U;
								m_pPelRec[2][curPosC] = dictInfo.pixel.V;
							}
						}
						//
						dictInfo.MatchType = MATCH_POS_ONE; uiSRBPosPrev=dictInfo.srbIndex; MatchTypePrev=MATCH_POS_ONE;
					}
					else
					{
						dictInfo.srbIndex = m_MaxPLTSize;
						m_bLoRefFlag=true;
						// reconstruction          
						UInt ydata = dictInfo.pixel.Y;
						UInt udata = dictInfo.pixel.U;
						UInt vdata = dictInfo.pixel.V;						
						//assert(dictInfo.length==1);
						for (UInt pixelCount = 0; pixelCount < dictInfo.length; pixelCount++)
						{
							Int curPosX;
							Int curPosY;
							Int curPixelPos = puiScanOrder[processedCount + pixelCount];
							TravX=getTravX(curPixelPos, CbSize    );
							TravY=getTravY(curPixelPos, log2CbSize);							
							if( isHorizontalScanning )
							{
								curPosX = xStartInPic + TravX;//curPixelPos % width;
								curPosY = yStartInPic + TravY;//curPixelPos / width;
							}
							else
							{
								curPosX = xStartInPic + TravY;//curPixelPos / height;
								curPosY = yStartInPic + TravX;//curPixelPos % height;
							}
							Int curPos = curPosY * m_strideRec[0] + curPosX;
							Int curPosC= (curPosY>>uiScaleY) * m_strideRec[1] + (curPosX>>uiScaleX);
							m_pPelRec[0][curPos ] = ydata;
							//m_pPelRec[1][curPosC] = udata;
							//m_pPelRec[2][curPosC] = vdata;
							if(!((curPosX & uiScaleX) || (curPosY & uiScaleY)) )
							{
								m_pPelRec[1][curPosC] = udata;
								m_pPelRec[2][curPosC] = vdata;
							}
						}
						dictInfo.MatchType = MATCH_NONE; uiSRBPosPrev=dictInfo.srbIndex; MatchTypePrev=MATCH_NONE;
					}
				}
				//add to hash
				addPixelToDict(dictInfo.length, true, xStartInPic, yStartInPic, width, height, processedCount, isHorizontalScanning);
				vecDictInfo.push_back( dictInfo );
				processedCount += dictInfo.length;
			}

			DictRDCost += tmpDictRDCost;
			if( DictRDCost > curBestRDCost)
				return false;
		}	 
	return true;
}

Void TComPseudoDict::decode( TComDataCU* pcCU, Int width, Int height, Int xStartInPic, Int yStartInPic, const vector<TDict1DInformation>& vecDictInfo, TCombinePixel* pPixels, Bool isHorizontalScanning )
{
	Int processedCount = 0;
	Int count = width*height;
	Int  CbSize    = width;
	Int  log2CbSize=g_aucConvertToBit[width]  + 2;
	UInt* puiScanOrder = g_scanOrder[SCAN_UNGROUPED][SCAN_TRAV][g_aucConvertToBit[width] + 2][g_aucConvertToBit[height] + 2];
	UInt uiIdxTrav;
  UInt uiScaleX = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);
	UInt uiScaleY = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
	//ChromaFormat chFormat = pcCU->getPic()->getChromaFormat();
	if ( isHorizontalScanning )
	{
		for ( Int i=0; i<vecDictInfo.size(); i++ )
		{
			if ( vecDictInfo[i].isMatched )
			{
				Int length = vecDictInfo[i].length;
				Int offsetX = vecDictInfo[i].offsetX;
				Int offsetY = vecDictInfo[i].offsetY;			
				for ( Int j=0; j<length; j++ )
				{
					uiIdxTrav = puiScanOrder[processedCount];
					Int currX = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
					Int currY = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;					
					Int curPos = (yStartInPic + currY)*m_strideRec[0] + (xStartInPic + currX);
					Int curPosC= ((yStartInPic + currY)>>uiScaleY)*m_strideRec[1] + ((xStartInPic + currX)>>uiScaleX);
					Int refPos = (yStartInPic + currY - offsetY)*m_strideRec[0] + (xStartInPic + currX - offsetX);
					Int refPosC= ((yStartInPic + currY - offsetY)>>uiScaleY)*m_strideRec[1] + ((xStartInPic + currX - offsetX)>>uiScaleX);
					m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
					if(!((currX & uiScaleX) || (currY & uiScaleY)) )
					{
						m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
						m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
					}
					pPixels[uiIdxTrav].Y = m_pPelRec[0][curPos ];
					pPixels[uiIdxTrav].U = m_pPelRec[1][curPosC];
					pPixels[uiIdxTrav].V = m_pPelRec[2][curPosC];
					processedCount++;
				}
			}
			else
			{
				uiIdxTrav = puiScanOrder[processedCount];
				Int currX = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
				Int currY = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;
				Int currOffsetX = xStartInPic + currX;
				Int currOffsetY = yStartInPic + currY;
				Int length = vecDictInfo[i].length;
				if( vecDictInfo[i].MatchType == MATCH_POS_WIDTH )
				{
					// printf("\nMATCH_POS_WIDTH");
					for ( Int j=0; j<length; j++ )
					{
						uiIdxTrav = puiScanOrder[processedCount];
						currX = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
						currY = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;
						currOffsetX = xStartInPic + currX;
						currOffsetY = yStartInPic + currY;
						Int curPos = currOffsetY*m_strideRec[0] + currOffsetX;
						Int curPosC= (currOffsetY>>uiScaleY)*m_strideRec[1] + (currOffsetX>>uiScaleX);
						Int refPos = (currOffsetY - 1)*m_strideRec[0] + (currOffsetX - 0);
						Int refPosC= ((currOffsetY - 1)>>uiScaleY)*m_strideRec[1] + ((currOffsetX - 0)>>uiScaleX);
					
						uiIdxTrav = currX + currY * width;
						m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
						if(!((currX & uiScaleX) || (currY & uiScaleY)) )
						{
							m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
							m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
						}
						pPixels[uiIdxTrav].Y = m_pPelRec[0][curPos ];
						pPixels[uiIdxTrav].U = m_pPelRec[1][curPosC];
						pPixels[uiIdxTrav].V = m_pPelRec[2][curPosC];
						processedCount++;
					}
				}
				else
				{
					for ( Int j=0; j<length; j++ )
					{
						uiIdxTrav = puiScanOrder[processedCount];
						currX = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
						currY = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;
						currOffsetX = xStartInPic + currX;
						currOffsetY = yStartInPic + currY;
						Int curPos = currOffsetY*m_strideRec[0] + currOffsetX;
						Int curPosC= (currOffsetY>>uiScaleY)*m_strideRec[1] + (currOffsetX>>uiScaleX);
						m_pPelRec[0][curPos] = vecDictInfo[i].pixel.Y;
						//m_pPelRec[1][curPosC] = vecDictInfo[i].pixel.U;
						//m_pPelRec[2][curPosC] = vecDictInfo[i].pixel.V;
						if(!((currX & uiScaleX) || (currY & uiScaleY)) )
						{
						m_pPelRec[1][curPosC] = vecDictInfo[i].pixel.U;
						m_pPelRec[2][curPosC] = vecDictInfo[i].pixel.V;
						}
						pPixels[uiIdxTrav].Y = m_pPelRec[0][curPos ];
						pPixels[uiIdxTrav].U = m_pPelRec[1][curPosC];
						pPixels[uiIdxTrav].V = m_pPelRec[2][curPosC];
						processedCount++;
					}
				}
			}
		}
	}
	else
	{
		for ( Int i=0; i<vecDictInfo.size(); i++ )
		{
			if ( vecDictInfo[i].isMatched )
			{
				Int length = vecDictInfo[i].length;
				Int offsetX = vecDictInfo[i].offsetX;
				Int offsetY = vecDictInfo[i].offsetY;

				for ( Int j=0; j<length; j++ )
				{
					uiIdxTrav = puiScanOrder[processedCount];
					Int currX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
					Int currY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
					Int curPos = (yStartInPic + currY)*m_strideRec[0] + (xStartInPic + currX);
					Int curPosC= ((yStartInPic + currY)>>uiScaleY)*m_strideRec[1] + ((xStartInPic + currX)>>uiScaleX);
					Int refPos = (yStartInPic + currY - offsetY)*m_strideRec[0] + (xStartInPic + currX - offsetX);
					Int refPosC= ((yStartInPic + currY - offsetY)>>uiScaleY)*m_strideRec[1] + ((xStartInPic + currX - offsetX)>>uiScaleX);
					
					uiIdxTrav = currX + currY * width;
					m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
					if(!((currX & uiScaleX) || (currY & uiScaleY)) )
					{
						m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
						m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
					}
					pPixels[uiIdxTrav].Y = m_pPelRec[0][curPos ];
					pPixels[uiIdxTrav].U = m_pPelRec[1][curPosC];
					pPixels[uiIdxTrav].V = m_pPelRec[2][curPosC];
					processedCount++;
				}
			}
			else
			{
				uiIdxTrav = puiScanOrder[processedCount];
				Int currX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
				Int currY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
				Int currOffsetX = xStartInPic + currX;
				Int currOffsetY = yStartInPic + currY;
				Int length = vecDictInfo[i].length;
				if( vecDictInfo[i].MatchType == MATCH_POS_WIDTH )
				{
					for ( Int j=0; j<length; j++ )
					{
						uiIdxTrav = puiScanOrder[processedCount];
						currX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
						currY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
						currOffsetX = xStartInPic + currX;
						currOffsetY = yStartInPic + currY;
						Int curPos = currOffsetY*m_strideRec[0] + currOffsetX;
						Int curPosC= (currOffsetY>>uiScaleY)*m_strideRec[1] + (currOffsetX>>uiScaleX);
						Int refPos = (currOffsetY - 0)*m_strideRec[0] + (currOffsetX - 1);
						Int refPosC= ((currOffsetY - 0)>>uiScaleY)*m_strideRec[1] + ((currOffsetX - 1)>>uiScaleX);
					
						uiIdxTrav = currX + currY * width;
						m_pPelRec[0][curPos ] = m_pPelRec[0][refPos ];
						if(!((currX & uiScaleX) || (currY & uiScaleY)) )
						{
							m_pPelRec[1][curPosC] = m_pPelRec[1][refPosC];
							m_pPelRec[2][curPosC] = m_pPelRec[2][refPosC];
						}
						pPixels[uiIdxTrav].Y = m_pPelRec[0][curPos ];
						pPixels[uiIdxTrav].U = m_pPelRec[1][curPosC];
						pPixels[uiIdxTrav].V = m_pPelRec[2][curPosC];
						processedCount++;
					}
				}
				else
				{
					for ( Int j=0; j<length; j++ )
					{
						uiIdxTrav = puiScanOrder[processedCount];
						currX = getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
						currY = getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
						currOffsetX = xStartInPic + currX;
						currOffsetY = yStartInPic + currY;
						Int curPos = currOffsetY*m_strideRec[0] + currOffsetX;
						Int curPosC= (currOffsetY>>uiScaleY)*m_strideRec[1] + (currOffsetX>>uiScaleX);
						m_pPelRec[0][curPos] = vecDictInfo[i].pixel.Y;
						//m_pPelRec[1][curPosC] = vecDictInfo[i].pixel.U;
						//m_pPelRec[2][curPosC] = vecDictInfo[i].pixel.V;
						if(!((currX & uiScaleX) || (currY & uiScaleY)) )
						{
						m_pPelRec[1][curPosC] = vecDictInfo[i].pixel.U;
						m_pPelRec[2][curPosC] = vecDictInfo[i].pixel.V;
						}
						uiIdxTrav = currX + currY * width;
						pPixels[uiIdxTrav].Y = m_pPelRec[0][curPos ];
						pPixels[uiIdxTrav].U = m_pPelRec[1][curPosC];
						pPixels[uiIdxTrav].V = m_pPelRec[2][curPosC]; 
						processedCount++;
					}
				}
			}
		}
	}

	assert( processedCount == count );
}

Void TComPseudoDict::addPixelToDict(Int count, Bool includeHash, Int xStartInPic, Int yStartInPic, Int width, Int height, Int iStartIdx, Bool bHorizontal )
{
	UInt* puiScanOrder = g_scanOrder[SCAN_UNGROUPED][SCAN_TRAV][g_aucConvertToBit[width] + 2][g_aucConvertToBit[height] + 2];
	Int  CbSize    = width;
	Int  log2CbSize=g_aucConvertToBit[width]  + 2;
	if (includeHash)
	{
		for (Int pixelCount = 0; pixelCount < count; pixelCount++)
		{
			Int currentX;// = xStartInPic + (iStartIdx + pixelCount) % width;
			Int currentY;// = yStartInPic + (iStartIdx + pixelCount) / width;
			UInt  uiIdx = iStartIdx + pixelCount;
			UInt  uiIdxTrav = puiScanOrder[uiIdx];
			if( bHorizontal )
			{
				currentX = xStartInPic + getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % width;
				currentY = yStartInPic + getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / width;
			}
			else
			{
				currentX = xStartInPic + getTravY(uiIdxTrav, log2CbSize);//uiIdxTrav / height;
				currentY = yStartInPic + getTravX(uiIdxTrav, CbSize    );//uiIdxTrav % height;
				uiIdxTrav= (currentY - yStartInPic) * width + (currentX - xStartInPic);
			}						
			//Int length = 1;
			Int lengthMinus1 = 0;
			Int startPositionX = currentX - lengthMinus1;
			Int startPositionY = currentY;
			if (startPositionX < 0 || startPositionY < 0)
			{
				continue;
			}

			UShort hashValue = (this->*getCRCValue)(m_pOrgPixelBuffer[uiIdxTrav]);//ZF


			m_dict[currentY][currentX].x = m_hashTable[hashValue].x;
			m_dict[currentY][currentX].y = m_hashTable[hashValue].y;
			m_hashTable[hashValue].x = startPositionX;
			m_hashTable[hashValue].y = startPositionY;
		}
	}
}
Double TComPseudoDict::getVerDictCost(Int offsetX, Int offsetY, UInt distortion,UInt offsetInCU,Bool isBoundaryCU,UInt cuWidth,UInt matchLength,UInt offsetSign,UInt leftPixelInCU)
{
	UInt sad = distortion;
	UInt numOfBits = 0; 
	//UInt mappedOffset = offsetInCU;
	// a bit for sign
	if(offsetX==1 && offsetY==0)
		numOfBits++;
	else
	{ 
		if(offsetY==0 && offsetX>1)
			offsetX--;	
		numOfBits++;
		//
		numOfBits++;
		if ( offsetY !=0 )
		{
			numOfBits++; 
			numOfBits += getOffsetBits( abs( offsetY ) - 1, 2 );
			if(offsetY<0)
			{

				if(offsetSign==1)
				{
					numOfBits += getOffsetBits( abs( offsetX )+1, 2 );
				}
				else
				{
					numOfBits += getOffsetBits( abs( offsetX )-1, 2 );
				}

			}
			else
			{
				numOfBits++; 
				if ( offsetX != 0 )
				{
					numOfBits++; 
					numOfBits += getOffsetBits( abs( offsetX )-1, 2 );
				}
			}
		}
		else
		{
			numOfBits += getOffsetBits( abs( offsetX )-1, 2 );
		}

	}

	if(matchLength!=0)
		numOfBits += getLenBits(matchLength-1,leftPixelInCU);  
	return this->getRdCostImp()->calcRdCost(numOfBits,sad);

}

Double TComPseudoDict::getDictCost(Int offsetX, Int offsetY, UInt distortion,UInt offsetInCU,Bool isBoundaryCU,UInt cuWidth,UInt matchLength,UInt offsetSign,UInt leftPixelInCU	)
{
	UInt sad = distortion;
	UInt numOfBits = 0; 
	UInt mappedOffset = offsetInCU;
	// a bit for sign
	if(offsetX==0 && offsetY==1)
		numOfBits++;
	else
	{ 
		if(offsetX==0 && offsetY>1)
			offsetY--;
		numOfBits++;
		numOfBits++;
		if ( offsetY !=0 )
		{
			if(!isBoundaryCU) numOfBits++; 
			numOfBits += getOffsetBits( abs( offsetY ) - 1, 2 );
			if(offsetY<0)
			{
				numOfBits+= getOffsetBits( abs( offsetX )-mappedOffset-1, 2 );
			}
			else
			{
				numOfBits++; 
				if ( offsetX != 0 )
				{
					numOfBits++; 
					numOfBits += getOffsetBits( abs( offsetX )-1, 2 );
				}
			}
		}
		else
		{
			if( offsetSign == 1 )
				numOfBits++;
			numOfBits += getOffsetBits( abs( offsetX )-1, 2 );

		}
	}
	if(matchLength!=0)
		numOfBits += getLenBits(matchLength-1,leftPixelInCU);  
	return this->getRdCostImp()->calcRdCost(numOfBits,sad);
}

UInt TComPseudoDict::getTBbits(Int uiSymbol,UInt uiMaxSymbol)
{
	UInt bins = 0;
	UInt uiThresh;
	if (uiMaxSymbol > 256)
	{
		UInt uiThreshVal = 1 << 8;
		uiThresh = 8;
		while (uiThreshVal <= uiMaxSymbol)
		{
			uiThresh++;
			uiThreshVal <<= 1;
		}
		uiThresh--;
	}
	else
	{
		uiThresh = g_uhPLTTBC[uiMaxSymbol];
	}

	UInt uiVal = 1 << uiThresh;
	assert(uiVal <= uiMaxSymbol);
	assert((uiVal << 1) > uiMaxSymbol);
	assert(uiSymbol < uiMaxSymbol);
	UInt b = uiMaxSymbol - uiVal;
	assert(b < uiVal);
	if (uiSymbol < uiVal - b)
	{
		bins = uiThresh;
	}
	else
	{
		bins = uiThresh+1;
	}
	return bins;

}

UInt TComPseudoDict::getLenBits(UInt len,UInt totalLeftPixel)
{
	UInt bins = 0;
	// printf("\n%d\t%d\t",len,totalPixel);
	assert( len <= totalLeftPixel );
	if ( len < 4 ) // 2 bits
	{
		if ( totalLeftPixel >= 4 )
		{
			bins++;
		}
		if(totalLeftPixel<3)
			bins += getTBbits(len,totalLeftPixel+1);
		else
			bins += 2;
	}
	else if ( len - 4 < 16 ) // 4 bits
	{
		bins +=1;
		if ( totalLeftPixel - 4 >= 16 )
		{
			bins+=1;
		}
		if(totalLeftPixel<19)
			bins += getTBbits(len-4,totalLeftPixel-3);
		else
			bins += 4;
	}
	else if ( len - 4 - 16 < 256 ) // 8 bits
	{
		bins +=1;
		bins +=1;
		if ( totalLeftPixel - 4 - 16 >= 256 )
		{
			bins +=1;
		}
		if(totalLeftPixel<275)
			bins += getTBbits(len-20,totalLeftPixel-19);
		else
			bins += 8;
	}
	else if ( len - 4 - 16 - 256 < 65536 ) // 12 bits, maximum number 4096
	{
		bins +=3;
		if(totalLeftPixel<4096)
			bins += getTBbits(len-276,totalLeftPixel-275);
		else
			bins += 12;
	}
	// printf("\n%d\t%d\t%d\t",len,totalPixel,bins);
	return bins;

}

UInt TComPseudoDict::getOffsetBits(Int uiSymbol,UInt uiCount)
{
	UInt bins = 0;
	Int numBins = 0;

	while( uiSymbol >= (UInt)(1<<uiCount) )
	{
		bins = 2 * bins + 1;
		numBins++;
		uiSymbol -= 1 << uiCount;
		uiCount  ++;
	}
	bins = 2 * bins + 0;
	numBins++;

	bins = (bins << uiCount) | uiSymbol;
	numBins += uiCount;  

	return numBins;

}

TComRdCost* TComPseudoDict::getRdCostImp()
{
	return this->m_pRdCostImp;

}
Void  TComPseudoDict::setRdCostImp (  TComRdCost* rdcImp)
{
	this->m_pRdCostImp = rdcImp;

}

Void TComPseudoDict::nonMajorColorQuantization( TComDataCU* pcCU, Pel orgData[3],UChar cqData[3], UInt uiMaxVal[3])
{
#if !ENABLE_ISC_SRB_QUANTIZATION
  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
    cqData[ch] = (UChar) orgData[ch];
  return ;
#endif
  Bool bLossless = pcCU->getCUTransquantBypass (0);
  if(bLossless)
	{
		for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
			cqData[ch] = (UChar) orgData[ch];
		return ;
	}
	Pel cqDataTemp=0,cqDataTemp1=0;

	Int iQP[3];
	Int iQPrem[3];
	Int iQPper[3];
	Int quantiserScale[3];
	Int quantiserRightShift[3];
	Int rightShiftOffset[3];
	Int InvquantiserRightShift[3];
	Int iAdd[3];
	for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
	{
    QpParam cQP(*pcCU, ComponentID(ch));
		iQP[ch] = cQP.Qp;
		iQPrem[ch] = iQP[ch] % 6;
		iQPper[ch] = iQP[ch] / 6;
		quantiserScale[ch] = g_quantScales[iQPrem[ch]];
		quantiserRightShift[ch] = QUANT_SHIFT + iQPper[ch];
		rightShiftOffset[ch] = 1 << (quantiserRightShift[ch] - 1);
		InvquantiserRightShift[ch] = IQUANT_SHIFT;
    iAdd[ch] = 1 << (InvquantiserRightShift[ch] - 1);

		cqDataTemp =  Pel(Clip3<Int>( 0, uiMaxVal[ch], ((orgData[ch] * quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch]) ));
		cqDataTemp1 = (((cqDataTemp *g_invQuantScales[iQPrem[ch]])<<iQPper[ch]) + iAdd[ch])>>InvquantiserRightShift[ch];
		cqData[ch] =(UChar)(ClipBD<Int>(cqDataTemp1, m_bitDepth[ch? 1:0]));  
	}
}

UInt TComPseudoDict::getAbsOffsetTExGolombBits(Int uiSymbol,UInt uiCount,UInt uiMaxValue)
{
	UInt bins = 0;
	Int numBins = 0;	
	UInt jieduan = 1<<uiCount;
	while( uiSymbol >= (UInt)(1<<uiCount) )
	{	
		bins = 2 * bins + 1;
		numBins++;
		uiSymbol -= 1 << uiCount;
		uiCount  ++;
		jieduan+=1<<uiCount;
	}	
	bins = 2 * bins + 0;	
	numBins++;
	if(uiMaxValue<jieduan)
	{	
		if(uiCount>1)
			uiCount--;	
	}
	numBins += uiCount;

	assert( numBins <= 32 ); 
	return numBins;
}

#if ISC_WPP_STEP
Bool TComPseudoDict::isValidIntraSCSearchArea( TComDataCU* pcCU,  Int iRefXInPic, Int iRefYInPic )
{
	const Int  cuPelX        = pcCU->getCUPelX() ;
	const Int  cuPelY        = pcCU->getCUPelY() ;
	Int uiRefCuX             = iRefXInPic>>m_uiCuWidthBits;
	Int uiRefCuY             = iRefYInPic>>m_uiCuHeightBits;
	Int uiCuPelX             = cuPelX >> m_uiCuWidthBits;
	Int uiCuPelY             = cuPelY >> m_uiCuHeightBits;
#if ISC_WPP_STEP == 1
	if((uiRefCuX - uiCuPelX) > (uiCuPelY - uiRefCuY))
#else
	if((uiRefCuX - uiCuPelX) > (uiCuPelY - uiRefCuY)*ISC_WPP_STEP)
#endif
	{
		return false;
	}
	return true;
}
#endif
#endif
