#pragma once

#include "TypeDef.h"
#include "TComRom.h"
#include "TComRdCost.h"
#include <vector>
using namespace std;

#if CE10_INTRA_STRING_COPY

#define MAX_DICT_MODE_DEPTH 4
#define MAX_DICT_TEST_CASE  1000
#define getTravX(uiTravIdx, CbSize)     ( uiTravIdx & (CbSize-1))
#define getTravY(uiTravIdx, log2CbSize) ( uiTravIdx >> log2CbSize)
enum ISC_TJ_MatchType
{
  MATCH_POS_WIDTH  = 0,
  MATCH_POS_ONE    = 1,
  MATCH_NONE       = 2,
  MATCH_PRB        = 3,
  NUMBER_OF_MATCH_TYPES = 4
};

enum DICT_STATUS
{
  DICT_STATUS_REC_HORIZONTAL,
  DICT_STATUS_REC_VERTICAL,
  DICT_STATUS_REC_BOTH,
};

enum DICT_CTX_IDX
{
  DICT_CURR_BEST,
  DICT_NEXT_BEST,
  DICT_TEMP_BEST,
  DICT_CTX_NUM,
};
struct TCombinePixel
{
  Pel Y;
  Pel U;
  Pel V;
};

Bool operator == (const TCombinePixel& _left, const TCombinePixel& _right);
Bool operator != (const TCombinePixel& _left, const TCombinePixel& _right);

struct TDict1DInformation
{
  Bool isMatched;
  Int length;
  Int offsetX;  // for psuedo 2D
  Int offsetY;  // for psuedo 2D
  TCombinePixel pixel;
  ISC_TJ_MatchType MatchType;
  UInt srbIndex;
};

#define DICT_PSEUDO_2D_HASH_BITS  12
#define DICT_PSEUDO_2D_HASH_SIZE  (1<<DICT_PSEUDO_2D_HASH_BITS)

#define ISC_UV_DISTORTION_WEIGTH               1
#if ISC_UV_DISTORTION_WEIGTH
#if STRING_COPY_RANGE == 2
const double lumaDistWeight = 1.0938;//1.0938;//1.125; 
const double chromaDistWeight = 0.938;//0.938;//0.938;
#elif  STRING_COPY_RANGE == 3
const double lumaDistWeight = 1.125;//1.0938;//1.125; 
const double chromaDistWeight = 0.938;//0.938;//0.938;
#elif  STRING_COPY_RANGE == 1
const double lumaDistWeight = 1.0;//625;//1.0938;//1.125; 
const double chromaDistWeight = 1.0;//0.875;//0.938;
#endif
#endif
struct TDictPreviousMatchPosition
{
  Short x;
  Short y;
};

struct TDictStatus
{
  TDictPreviousMatchPosition   m_hashTable[DICT_PSEUDO_2D_HASH_SIZE];
  TDictPreviousMatchPosition   m_dict[MAX_CU_SIZE][MAX_CU_SIZE];
  Int m_currDictSize;

  Void save(TDictStatus* p);
};

class TComPicYuv;
class TComDataCU;

class TComPseudoDict
{
public:
  TComPseudoDict( Int picWidth, Int picHeight, UInt uiMaxCUWidth, UInt uiMaxCUHeight, Bool isEncoding = true );
  ~TComPseudoDict();

private:
  TComPseudoDict( const TComPseudoDict& comPseudoDict );
  const TComPseudoDict& operator = (const TComPseudoDict& comPseudoDict);
  TComRdCost*             m_pRdCostImp;
public:
  Void reset( TComPicYuv* pOrg, TComPicYuv* pRec );
  Void addToDict( Int xStartInPic, Int yStartInPic, Int width, Int height, UInt uiScaleX, UInt uiScaleY );
  Bool encode(TComDataCU* pcCU, Int width, Int height, Int xStartInPic, Int yStartInPic, vector<TDict1DInformation>& vecDictInfo, Double curBestRDCost );
  Void decode(TComDataCU* pcCU, Int width, Int height, Int xStartInPic, Int yStartInPic, const vector<TDict1DInformation>& vecDictInfo, TCombinePixel* pPixels, Bool isHorizontalScanning );
  Void addPixelToDict(Int count, Bool includeHash, Int xStartInPic, Int yStartInPic, Int width, Int height, Int iStartIdx, Bool bHorizontal=true );
  
  Void setBitDepth (ChannelType type, Int u ) { m_bitDepth[type]=u; };
  TComRdCost*            getRdCostImp();
  Void  setRdCostImp (  TComRdCost* rdcImp);
  Double getDictCost(Int offsetX, Int offsetY, UInt distortion,UInt offsetInCU,Bool isBoundaryCU,UInt cuWidth,UInt matchLength,UInt offsetSign,UInt leftPixelInCU );
  UInt   getLenBits(UInt len,UInt totalPixel);
 /* UInt   getMTKLenBits(UInt len,UInt totalLeftPixel);*/
  UInt   getOffsetBits(Int uiSymbol,UInt uiCount);
  UInt getAbsOffsetTExGolombBits(Int uiSymbol,UInt uiCount,UInt uiMax);
  Double getVerDictCost(Int offsetX, Int offsetY, UInt distortion,UInt offsetInCU,Bool isBoundaryCU,UInt cuWidth,UInt matchLength,UInt offsetSign,UInt leftPixelInCU );
 /* UInt   getTruncMsbP1Bits(UInt len,UInt uiMax,UInt& bits);*/
  UInt getTBbits(Int uiSymbol,UInt uiMaxSymbol);
  TCRCCalculatorLight m_crcCalculator;
  UShort xGetCRCValue( UChar* p, Int length );

  UShort getCRCValueLossy(TCombinePixel pixelCur); // calculating in reverse order, including the pixel at (xEnd, yEnd)
  UShort getCRCValueLossless(TCombinePixel pixelCur); // calculating in reverse order, including the pixel at (xEnd, yEnd)
  UShort (TComPseudoDict::*getCRCValue)(TCombinePixel);

  Void save(TDictStatus* pDictStatus, Int xLCUStart, Int yLCUStart, Int width, Int height);
  Void restore(TDictStatus* pDictStatus, Int xLCUStart, Int yLCUStart, Int width, Int height);
  //
  Void deriveCluster(TComDataCU* pcCU, UInt uiWidth, UInt uiHeight, Int xStartInPic, Int yStartInPic, Bool bLosslessMode, Double dLambda);
  Void nonMajorColorQuantization( TComDataCU* pcCU, Pel orgData[3],UChar cqData[3], UInt uiMaxVal[3]);
  Void saveSRB      (TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth);
  Void SRBPredition (TComDataCU* pcCU, TCombinePixel *pHiRefUmP, UInt uiHiRefUmPSize, UChar* pReusedPrev);
  Int  getSrbLength (TComDataCU*pcCU, Int width, Int height, Int xCuStartInPic, Int yCuStartInPic, Int offsetInCu, UInt& dist, TCombinePixel& pixelCur);
  Int  getSrbAboveLength (TComDataCU*pcCU, Int width, Int height, Int xCuStartInPic, Int yCuStartInPic, Int offsetInCu, UInt& dist, TCombinePixel& pixelCur, Bool bUseOrg=true);
  UInt getSrbBits   (TComDataCU*pcCU, TCombinePixel pixelCur, Int srbLength, UInt uiIdx, UInt uiTotal, UInt& uiSRBPos, UInt uiSRBPosPrev, ISC_TJ_MatchType MatchTypePrev);
  UInt CalcDist     (TCombinePixel orgPixel, TCombinePixel recPixel );
  Int EstWriteSRBPos(UInt uiIdx, UInt uiSRBPos, Int iMaxSymbol, UInt uiSRBPosPrev, ISC_TJ_MatchType MatchTypePrev);
  Int EstWriteTruncBinCode(UInt uiSymbol, UInt uiMaxSymbol);
  Int findPosInHiRefUmP( TCombinePixel pix, TCombinePixel *pHiRefUmP, UInt uiHiRefUmPSize)
  {
    Int pos = -1;
    for(Int i=0;i<uiHiRefUmPSize;i++)
    {
      if(pix == pHiRefUmP[i])
      {
        pos = i;
        break;
      }
    }
    return pos;
  };
  Void setMaxPredSize (UInt uiValue ) { m_PLTMaxPredSize=uiValue; };
  Void setMaxPltSize  (UInt uiValue ) { m_MaxPLTSize=uiValue; };
  Void setClusterPred (Bool bFlag   ) { m_PredFlag=bFlag; };

#if ISC_WPP_STEP
  inline  Bool isValidIntraSCSearchArea( TComDataCU* pcCU,  Int predX, Int predY );
#endif
private:
  TDictPreviousMatchPosition   m_hashTable[DICT_PSEUDO_2D_HASH_SIZE];
  TDictPreviousMatchPosition** m_dict;
  TDictStatus                  m_DictStatus;
  UInt m_uiMaxCuWidth;
  UInt m_uiMaxCuHeight;
#if ISC_WPP_STEP
  
  UInt m_uiCuWidthBits;
  UInt m_uiCuHeightBits;
#endif
  Int m_picWidth;
  Int m_picHeight;
  Bool m_isEncoding;
  Pel* m_pPelOrg[3];
  Pel* m_pPelRec[3];
  Int  m_strideOrg[3];
  Int  m_strideRec[3];
  UChar* m_pCRCCalculationBuffer;
  Int  m_bitDepth[MAX_NUM_CHANNEL_TYPE];
  UInt m_PLTMaxPredSize;
  UInt m_MaxPLTSize;
  Bool m_PredFlag;
  //
  Bool m_bMajor[MAX_CU_SIZE*MAX_CU_SIZE];
  UChar m_cMatchIndex[MAX_CU_SIZE*MAX_CU_SIZE];
  TCombinePixel m_pOrgPixelBuffer[MAX_CU_SIZE*MAX_CU_SIZE];
  TCombinePixel m_pCQPixelBuffer[MAX_CU_SIZE*MAX_CU_SIZE];
  Int  m_uiSRBSize;
  Int  m_uiHiRefUmPSize;
  Bool m_bHiRefFlag;
  Bool m_bLoRefFlag;
  TCombinePixel m_HiRefUmPixel[MAX_SRB_SIZE];
};

#endif
